# -*- coding: utf-8 -*-

"""Module for writting correlator output to a FITS IDI file.  The classes and 
functions defined in this module are based heavily off the lwda_fits library."""

import os
import gc
import re
import sys
import math
import numpy
import pyfits
from datetime import datetime

from lsl import astro
from lsl.common import constants
from lsl.common import dp as dp_common
from lsl.common.stations import geo2ecef
from lsl.correlator import uvUtils
from lsl.misc import mathutil
from lsl.misc import geodesy
from lsl.common.warns import warnExperimental

__version__ = '0.2'
__revision__ = '$ Revision: 12 $'
__all__ = ['IDI', 'StokesCodes', '__version__', '__revision__', '__all__']


IDIVersion = (2, 0)

StokesCodes = {'I': 1, 'Q': 2, 'U': 3, 'V': 4, 
			'RR': -1, 'LL': -2, 'RL': -3, 'LR': -4, 
			'XX': -5, 'YY': -6, 'XY': -7, 'YX': -8}


class IDI(object):
	"""Class for storing visibility data and writing the data, along with array
	geometry, frequency setup, etc., to a FITS IDI file that can be read into 
	AIPS via the FITLD task."""

	class _Antenna(object):
		"""Holds information describing the location and properties of an antenna."""

		def __init__(self, id, x, y, z, bits=8):
			self.id = id
			self.x = x
			self.y = y
			self.z = z
			self.levels = bits
			self.polA = {'Type': 'X', 'Angle': 0.0, 'Cal': 1.0}
			self.polB = {'Type': 'Y', 'Angle': 0.0, 'Cal': 1.0}

		def getName(self):
			return "LWA%03i" % self.id

	class _Frequency:
		"""Holds information about the frequency setup used in the file."""

		def __init__(self, offset, channelWidth, bandwidth):
			self.id = 1
			self.bandFreq = offset
			self.chWidth = channelWidth
			self.totalBW = bandwidth
			self.sideBand = 1
			self.baseBand = 0

	class _UVData(object):
		"""Represents one UV visibility data set for a given observation time."""
    
		def __init__(self, obsTime, intTime, dataDict):
			self.obsTime = obsTime
			self.intTime = intTime
			self.dataDict = dataDict
		
		def time(self):
			return self.obsTime

	def parseRefTime(self, refTime):
		"""Given a time as either a integer, float, string, or datetime object, 
		convert it to a string in the formation 'YYYY-MM-DDTHH:MM:SS'."""

		# Valid time string (modulo the 'T')
		timeRE = re.compile(r'\d{4}-\d{2}-\d{2}[ T]\d{2}:\d{2}:\d{2}(\.\d+)?')

		if type(refTime).__name__ in ['int', 'long', 'float']:
			refDateTime = datetime.utcfromtimestamp(refTime)
			refTime = refDateTime.strftime("%Y-%m-%dT%H:%M:%S")
		elif type(refTime).__name__ in ['datetime']:
			refTime = refTime.strftime("%Y-%m-%dT%H:%M:%S")
		elif type(refTime).__name__ in ['str']:
			# Make sure that the string times are of the correct format
			if re.match(timeRE, refTime) is None:
				raise RuntimeError("Malformed date/time provided: %s" % refTime)
			else:
				refTime = refTime.replace(' ', 'T', 1)
		else:
			raise RuntimeError("Unknown time format provided.")

		return refTime

	def refTime2AstroDate(self):
		"""Convert a reference time string to an :class:`lsl.astro.date` object."""

		dateStr = self.refTime.replace('T', '-').replace(':', '-').split('-')
		return astro.date(int(dateStr[0]), int(dateStr[1]), int(dateStr[2]), int(dateStr[3]), int(dateStr[4]), float(dateStr[5]))


	def __init__(self, filename, refTime=0.0, verbose=False):
		"""Initialize a new FITS IDI object using a filename and a reference time 
		given in seconds since the UNIX 1970 ephem, a python datetime object, or a 
		string in the format of 'YYYY-MM-DDTHH:MM:SS'."""

		# File-specific information
		self.filename = filename
		self.verbose = verbose

		# Observation-specifc information
		self.refTime = self.parseRefTime(refTime)
		self.nAnt = 0
		self.nChan = 0
		self.nStokes = 0
		self.refVal = 0
		self.refPix = 0
		self.channelWidth = 0

		# Parameters that store the meta-data and data
		self.array = []
		self.freq = []
		self.stokes = []
		self.data = []

		# Open the file and get going
		self.FITS = pyfits.open(filename, mode='append')

	def setStokes(self, polList):
		"""Given a list of Stokes parameters, update the object's parameters"""

		for pol in polList:
			if type(pol).__name__ == 'str':
				self.stokes.append(StokesCodes[pol.upper()])
			else:
				self.stokes.append(pol)

		self.nStokes = len(self.stokes)

	def setFrequency(self, freq):
		"""Given a numpy array of frequencies, set the relevant common observation
		parameters and add an entry to the self.freq list."""

		self.nChan = len(freq)
		self.refVal = freq[0]
		self.refPix = 1
		self.channelWidth = numpy.abs(freq[1] - freq[0])
		totalWidth = numpy.abs(freq[-1] - freq[0])

		freqSetup = self._Frequency(0.0, self.channelWidth, totalWidth)
		self.freq.append(freqSetup)

	def setGeometry(self, site, stands, bits=8):
		"""Given a station and an array of stands, set the relevant common observation
		parameters and add entries to the self.array list."""

		# Make sure that we have been passed 255 or fewer stands
		if len(stands) > 255:
			raise RuntimeError("FITS IDI support up to 255 antennas only, given %i" % len(stands))

		arrayX, arrayY, arrayZ = site.getGeocentricLocation()
		xyz = uvUtils.getXYZ(stands)

		# Create the stand mapper to deal with the fact that stands range from 
		# 1 to 258, not 1 to 255
		mapper = {}
		if stands.max() > 255:
			enableMapper = True
		else:
			enableMapper = False

		ants = []
		topo2eci = site.getECEFTransform()
		for i in range(len(stands)):
			eci = numpy.dot(topo2eci, xyz[i,:])
			ants.append( self._Antenna(stands[i], eci[0], eci[1], eci[2], bits=bits) )
			if enableMapper:
				mapper[stands[i]] = i+1
			else:
				mapper[stands[i]] = stands[i]

		# If the mapper has been enabled, tell the user about it
		if enableMapper and self.verbose:
			print "FITS IDI: stand ID mapping enabled"
			for key, value in mapper.iteritems():
				print "FITS IDI:  stand #%i -> mapped #%i" % (key, value)

		self.nAnt = len(ants)
		self.array.append( {'center': [arrayX, arrayY, arrayZ], 'ants': ants, 'mapper': mapper, 'enableMapper': enableMapper} )

	def addDataSet(self, obsTime, intTime, baselines, visibilities):
		"""Create a UVData object to store a collection of visibilites."""

		dataDict = {}
		for (stand1,stand2), visData in zip(baselines, visibilities):
			baseline = (stand1 << 16) | stand2
			dataDict[baseline] = visData
			
		self.data.append( self._UVData(obsTime, intTime, dataDict) )

	def write(self):
		"""Fill in the FITS-IDI file will all of the tables in the 
		correct order."""
		
		self.__writePrimary()
		self.__writeGeometry()
		self.__writeFrequency()
		self.__writeAntenna()
		self.__writeBandpass()
		self.__writeSource()
		self.__writeData()

		# Clear out the data section
		del(self.data[:])
		gc.collect()

	def __addCommonKeywords(self, hdr, name, revision):
		"""Added keywords common to all table headers."""

		hdr.update('EXTNAME', name, 'FITS-IDI table name')
		hdr.update('EXTVER', 1, 'table instance number') 
		hdr.update('TABREV', revision, 'table format revision number')
		hdr.update('NO_STKD', 1, 'number of Stokes parameters')
		hdr.update('STK_1', self.stokes[0], 'first Stokes parameter')
		hdr.update('NO_BAND', 1, 'number of frequency bands')
		hdr.update('NO_CHAN', self.nChan, 'number of frequency channels')
		hdr.update('REF_FREQ', self.refVal, 'reference frequency (Hz)')
		hdr.update('CHAN_BW', self.channelWidth, 'channel bandwidth (Hz)')
		hdr.update('REF_PIXL', float(self.refPix), 'reference frequency bin')
	
		date = self.refTime.split('-')
		name = "ZA%s%s%s" % (date[0][2:], date[1], date[2])
		hdr.update('OBSCODE', name, 'zenith all-sky image')
	
		hdr.update('ARRNAM', 'LWA-1')      
		hdr.update('RDATE', self.refTime, 'file data reference date')

	def __makeAppendTable(self, extension, AddRows=1):
		"""Private function to make a temporary table for appending data."""

		nrows = self.hdulist[extension].data.shape[0]
		tempHDU = pyfits.new_table(self.hdulist[extension].columns, nrows=nrows+AddRows)
		for key in list(self.hdulist[extension].header.keys()):
			tempHDU.header.update(key, self.hdulist[extension].header[key])
	
		return tempHDU

	def __applyAppendTable(self, extension, tempHDU):
		"""Private function to replace the given extension with the temporary
		table."""

		self.hdulist[extension] = tempHDU
		self.flush()

	def __writePrimary(self):
		"""Write the primary HDU to file."""

		primary = pyfits.PrimaryHDU()

		primary.header.update('NAXIS', 0, 'indicates IDI file')
		primary.header.update('EXTEND', True, 'indicates IDI file')
		primary.header.update('GROUPS', True, 'indicates IDI file')
		primary.header.update('GCOUNT', 0)
		primary.header.update('PCOUNT', 0)
		primary.header.update('OBJECT', 'BINARYTB')
		primary.header.update('TELESCOP', 'LWA-1')
		primary.header.update('INSTRUME', 'LWA-1')
		primary.header.update('OBSERVER', 'ZASKY', 'zenith all-sky image')
		primary.header.update('ORIGIN', 'LSL')
		primary.header.update('CORRELAT', 'LWASWC', 'Correlator used')
		primary.header.update('FXCORVER', '1', 'Correlator version')
		primary.header.update('LWATYPE', 'IDI-ZA', 'LWA FITS file type')
		primary.header.update('LWAMAJV', IDIVersion[0], 'LWA FITS file format major version')
		primary.header.update('LWAMINV', IDIVersion[1], 'LWA FITS file format minor version')
		primary.header.update('DATE-OBS', self.refTime, 'IDI file data collection date')
		ts = str(astro.get_date_from_sys())
		primary.header.update('DATE-MAP', ts.split()[0], 'IDI file creation date')
		
		self.FITS.append(primary)
		self.FITS.flush()

	def __writeGeometry(self):
		"""Define the Array_Geometry table (group 1, table 1)."""

		i = 0
		names = []
		xyz = numpy.zeros((self.nAnt,3), dtype=numpy.float64)
		for ant in self.array[0]['ants']:
			xyz[i,:] = numpy.array([ant.x, ant.y, ant.z])
			names.append(ant.getName())
			i = i + 1

		# Antenna name
		c1 = pyfits.Column(name='ANNAME', format='A8', 
						array=numpy.array([ant.getName() for ant in self.array[0]['ants']]))
		# Station coordinates in meters
		c2 = pyfits.Column(name='STABXYZ', unit='METERS', format='3D', 
						array=xyz)
		# First order derivative of station coordinates in m/s
		c3 = pyfits.Column(name='DERXYZ', unit='METERS/S', format='3E', 
						array=numpy.zeros((self.nAnt,3), dtype=numpy.float32))
		# Orbital elements
		c4 = pyfits.Column(name='ORBPARM', format='1D', 
						array=numpy.zeros((self.nAnt,), dtype=numpy.float64))
		# Station number
		c5 = pyfits.Column(name='NOSTA', format='1J', 
						array=numpy.array([self.array[0]['mapper'][ant.id] for ant in self.array[0]['ants']]))
		# Mount type (0 == alt-azimuth)
		c6 = pyfits.Column(name='MNTSTA', format='1J', 
						array=numpy.zeros((self.nAnt,), dtype=numpy.int32))
		# Axis offset in meters
		c7 = pyfits.Column(name='STAXOF', unit='METERS', format='3E', 
						array=numpy.zeros((self.nAnt,3), dtype=numpy.float32))

		# Define the collection of columns
		colDefs = pyfits.ColDefs([c1, c2, c3, c4, c5, c6, c7])

		# Create the table and fill in the header
		ag = pyfits.new_table(colDefs)
		self.__addCommonKeywords(ag.header, 'ARRAY_GEOMETRY', 1)

		ag.header.update('EXTVER', 1, 'array ID')
		ag.header.update('ARRNAM', 'LWA-1')
		ag.header.update('FRAME', 'GEOCENTRIC', 'coordinate system')
		ag.header.update('NUMORB', 0, 'number of orbital parameters')
		ag.header.update('FREQ', self.refVal, 'reference frequency (Hz)')
		ag.header.update('TIMSYS', 'UTC', 'time coordinate system')

		date = self.refTime2AstroDate()
		utc0 = date.to_jd()
		gst0 = astro.get_apparent_sidereal_time(utc0)
		ag.header.update('GSTIA0', gst0 * 15, 'GAST (deg) at RDATE 0 hours')
		
		utc1 = utc0 + 1
		gst1 = astro.get_apparent_sidereal_time(utc1)
		if gst1 < gst0:
			gst1 += 24.0
		ds = gst1 - gst0
		deg = ds * 15.0      
		ag.header.update('DEGPDY', 360.0 + deg, 'rotation rate of the earth (deg/day)')
		
		refDate = self.refTime2AstroDate()
		refMJD = refDate.to_jd() - astro.MJD_OFFSET
		eop = geodesy.getEOP(refMJD)
		if eop is None:
			eop = [geodesy.EOP(mjd=refMJD)]

		ag.header.update('UT1UTC', eop[0].utDiff, 'difference UT1 - UTC for reference date')
		ag.header.update('IATUTC', astro.leap_secs(utc0), 'TAI - UTC for reference date')
		ag.header.update('POLARX', eop[0].x)
		ag.header.update('POLARY', eop[0].y)

		ag.header.update('ARRAYX', self.array[0]['center'][0], 'array ECI X coordinate (m)')
		ag.header.update('ARRAYY', self.array[0]['center'][1], 'array ECI Y coordinate (m)')
		ag.header.update('ARRAYZ', self.array[0]['center'][2], 'array ECI Z coordinate (m)')

		ag.header.update('NOSTAMAP', int(self.array[0]['enableMapper']), 'Mapping enabled for stand numbers')

		ag.name = 'ARRAY_GEOMETRY'
		self.FITS.append(ag)
		self.FITS.flush()

		if self.array[0]['enableMapper']:
			self.__writeMapper()

	def __writeFrequency(self):
		"""Define the Frequency table (group 1, table 3)."""

		# Frequency setup number
		c1 = pyfits.Column(name='FREQID', format='1J', 
						array=numpy.array([self.freq[0].id], dtype=numpy.int32))
		# Frequency offsets in Hz
		c2 = pyfits.Column(name='BANDFREQ', format='1D', unit='HZ', 
						array=numpy.array([self.freq[0].bandFreq], dtype=numpy.float64))
		# Channel width in Hz
		c3 = pyfits.Column(name='CH_WIDTH', format='1E', unit='HZ', 
						array=numpy.array([self.freq[0].chWidth], dtype=numpy.float32))
		# Total bandwidths of bands
		c4 = pyfits.Column(name='TOTAL_BANDWIDTH', format='1E', unit='HZ', 
						array=numpy.array([self.freq[0].totalBW], dtype=numpy.float32))
		# Sideband flag
		c5 = pyfits.Column(name='SIDEBAND', format='1J', 
						array=numpy.array([self.freq[0].sideBand], dtype=numpy.int32))
		# Baseband channel
		c6 = pyfits.Column(name='BB_CHAN', format='1J', 
						array=numpy.array([self.freq[0].baseBand], dtype=numpy.int32))

		# Define the collection of columns
		colDefs = pyfits.ColDefs([c1, c2, c3, c4, c5, c6])

		# Create the table and header
		fq = pyfits.new_table(colDefs)
		self.__addCommonKeywords(fq.header, 'FREQUENCY', 1)
		
		# Add the table to the file
		fq.name = 'FREQUENCY'
		self.FITS.append(fq)
		self.FITS.flush()

	def __writeAntenna(self):
		"""Define the Antenna table (group 2, table 1)."""

		# Central time of period covered by record in days
		c1 = pyfits.Column(name='TIME', unit='DAYS', format='1D', 
						array=numpy.zeros((self.nAnt,), dtype=numpy.float64))
		# Durration of period covered by record in days
		c2 = pyfits.Column(name='TIME_INTERVAL', unit='DAYS', format='1E', 
						array=(2*numpy.ones((self.nAnt,), dtype=numpy.float32)))
		# Antenna name
		c3 = pyfits.Column(name='ANNAME', format='A8', 
						array=self.FITS['ARRAY_GEOMETRY'].data.field('ANNAME'))
		# Antenna number
		c4 = pyfits.Column(name='ANTENNA_NO', format='1J', 
						array=self.FITS['ARRAY_GEOMETRY'].data.field('NOSTA'))
		# Array number
		c5 = pyfits.Column(name='ARRAY', format='1J', 
						array=numpy.ones((self.nAnt,), dtype=numpy.int32))
		# Frequency setup number
		c6 = pyfits.Column(name='FREQID', format='1J', 
						array=(numpy.zeros((self.nAnt,), dtype=numpy.int32) + self.freq[0].id))
		# Number of digitizer levels
		c7 = pyfits.Column(name='NO_LEVELS', format='1J', 
						array=numpy.array([ant.levels for ant in self.array[0]['ants']]))
		# Feed A polarization label
		c8 = pyfits.Column(name='POLTYA', format='A1', 
						array=numpy.array([ant.polA['Type'] for ant in self.array[0]['ants']]))
		# Feed A orientation in degrees
		c9 = pyfits.Column(name='POLAA', format='1E', 
						array=numpy.array([ant.polA['Angle'] for ant in self.array[0]['ants']], dtype=numpy.float32))
		# Feed A polarization parameters
		c10 = pyfits.Column(name='POLCALA', format='1E', 
						array=numpy.array([ant.polA['Cal'] for ant in self.array[0]['ants']], dtype=numpy.float32))
		# Feed B polarization label
		c11 = pyfits.Column(name='POLTYB', format='A1', 
						array=numpy.array([ant.polB['Type'] for ant in self.array[0]['ants']]))
		# Feed B orientation in degrees
		c12 = pyfits.Column(name='POLAB', format='1E', 
						array=numpy.array([ant.polB['Angle'] for ant in self.array[0]['ants']], dtype=numpy.float32))
		# Feed B polarization parameters
		c13 = pyfits.Column(name='POLCALB', format='1E', 
						array=numpy.array([ant.polB['Cal'] for ant in self.array[0]['ants']], dtype=numpy.float32))

		colDefs = pyfits.ColDefs([c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, 
							c11, c12, c13])

		# Create the Antenna table and update it's header
		an = pyfits.new_table(colDefs)
		self.__addCommonKeywords(an.header, 'ANTENNA', 1)

		an.header.update('NOPCAL', 0, 'number of polarization parameters')
		an.header.update('POLTYPE', 'X-Y LIN', 'polarization parameterization')
		
		an.name = 'ANTENNA'
		self.FITS.append(an)
		self.FITS.flush()

	def __writeBandpass(self):
		"""Define the Bandpass table (group 2, table 3)."""

		# Central time of period covered by record in days
		c1 = pyfits.Column(name='TIME', unit='DAYS', format='1D', 
						array=numpy.zeros((self.nAnt,), dtype=numpy.float64))
		# Durration of period covered by record in days
		c2 = pyfits.Column(name='TIME_INTERVAL', unit='DAYS', format='1E',
						array=(2*numpy.ones((self.nAnt,), dtype=numpy.float32)))
		# Source ID
		c3 = pyfits.Column(name='SOURCE_ID', format='1J', 
						array=numpy.zeros((self.nAnt,), dtype=numpy.int32))
		# Antenna number
		c4 = pyfits.Column(name='ANTENNA_NO', format='1J', 
						array=self.FITS['ANTENNA'].data.field('ANTENNA_NO'))
		# Array number
		c5 = pyfits.Column(name='ARRAY', format='1J', 
						array=numpy.ones((self.nAnt,), dtype=numpy.int32))
		# Frequency setup number
		c6 = pyfits.Column(name='FREQID', format='1J',
						array=(numpy.zeros((self.nAnt,), dtype=numpy.int32) + self.freq[0].id))
		# Bandwidth in Hz
		c7 = pyfits.Column(name='BANDWIDTH', unit='HZ', format='1E',
						array=(numpy.zeros((self.nAnt,), dtype=numpy.float32)+self.freq[0].totalBW))
		# Band frequency in Hz
		c8 = pyfits.Column(name='BAND_FREQ', unit='HZ', format='1D',
						array=(numpy.zeros((self.nAnt,), dtype=numpy.float64)+self.freq[0].bandFreq))
		# Referance antenna number (pol. 1)
		c9 = pyfits.Column(name='REFANT_1', format='1J',
						array=numpy.ones((self.nAnt,), dtype=numpy.int32))
		# Real part of the bandpass (pol. 1)
		c10 = pyfits.Column(name='BREAL_1', format='%dE' % self.nChan,
						array=numpy.ones((self.nAnt,self.nChan), dtype=numpy.float32))
		# Imagniary part of the bandpass (pol. 1)
		c11 = pyfits.Column(name='BIMAG_1', format='%dE' % self.nChan,
						array=numpy.zeros((self.nAnt,self.nChan), dtype=numpy.float32))
		# Referance antenna number (pol. 2)
		c12 = pyfits.Column(name='REFANT_2', format='1J',
						array=numpy.ones((self.nAnt,), dtype=numpy.int32))
		# Real part of the bandpass (pol. 2)
		c13 = pyfits.Column(name='BREAL_2', format='%dE' % self.nChan,
						array=numpy.ones((self.nAnt,self.nChan), dtype=numpy.float32))
		# Imagniary part of the bandpass (pol. 2)
		c14 = pyfits.Column(name='BIMAG_2', format='%dE' % self.nChan,
						array=numpy.zeros((self.nAnt,self.nChan), dtype=numpy.float32))

		colDefs = pyfits.ColDefs([c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, 
							c11, c12, c13, c14])

		# Create the Bandpass table and update its header
		bp = pyfits.new_table(colDefs)
		self.__addCommonKeywords(bp.header, 'BANDPASS', 1)

		bp.header.update('NO_ANT', self.nAnt)
		bp.header.update('NO_POL', 2)
		bp.header.update('NO_BACH', self.nChan)
		bp.header.update('STRT_CHN', self.refPix)

		bp.name = 'BANDPASS'
		self.FITS.append(bp)
		self.FITS.flush()

	def __writeSource(self):
		"""Define the Source table (group 1, table 2)."""

		hrz = astro.hrz_posn(0, 90)
		(arrPos, ag) = self.readArrayGeometry()
		ids = ag.keys()

		nameList = []
		raList = []
		decList = []
		raPoList = []
		decPoList = []
		sourceID = 1
		for dataSet in self.data:
			utc = astro.taimjd_to_utcjd(dataSet.obsTime)
			date = astro.get_date(utc)
			date.hours = 0
			date.minutes = 0
			date.seconds = 0
			utc0 = date.to_jd()
			equ = hrz.to_equ(arrPos, utc)
			
			# current apparent zenith equatorial coordinates
			raList.append(equ.ra)
			decList.append(equ.dec)
			
			# J2000 zenith equatorial coordinates
			equPo = astro.get_equ_prec2(equ, utc, astro.J2000_UTC_JD)
			raPoList.append(equPo.ra)
			decPoList.append(equPo.dec)
			
			# format 'source' name based on local sidereal time
			raHms = astro.deg_to_hms(equ.ra)
			(tsecs, secs) = math.modf(raHms.seconds)
			name = "ZA%02d%02d%02d%01d" % (raHms.hours, raHms.minutes, int(secs), int(tsecs * 10.0))
			nameList.append(name)
			sourceID += 1
		nSource = len(nameList)

		# Source ID number
		c1 = pyfits.Column(name='SOURCE_ID', format='1J', 
						array=numpy.arange(1, nSource+1, dtype=numpy.int32))
		# Source name
		c2 = pyfits.Column(name='SOURCE', format='A16', 
						array=numpy.array(nameList))
		# Source qualifier
		c3 = pyfits.Column(name='QUAL', format='1J', 
						array=numpy.zeros((nSource,), dtype=numpy.int32))
		# Calibrator code
		c4 = pyfits.Column(name='CALCODE', format='A4', 
						array=numpy.array(('   ',)).repeat(nSource))
		# Frequency group ID
		c5 = pyfits.Column(name='FREQID', format='1J', 
						array=(numpy.zeros((nSource,), dtype=numpy.int32)+self.freq[0].id))
		# Stokes I flux density in Jy
		c6 = pyfits.Column(name='IFLUX', format='1E', 
						array=numpy.zeros((nSource,), dtype=numpy.float32))
		# Stokes I flux density in Jy
		c7 = pyfits.Column(name='QFLUX', format='1E', 
						array=numpy.zeros((nSource,), dtype=numpy.float32))
		# Stokes I flux density in Jy
		c8 = pyfits.Column(name='UFLUX', format='1E', 
						array=numpy.zeros((nSource,), dtype=numpy.float32))
		# Stokes I flux density in Jy
		c9 = pyfits.Column(name='VFLUX', format='1E', 
						array=numpy.zeros((nSource,), dtype=numpy.float32))
		# Spectral index
		c10 = pyfits.Column(name='ALPHA', format='1E', 
						array=numpy.zeros((nSource,), dtype=numpy.float32))
		# Frequency offset in Hz
		c11 = pyfits.Column(name='FREQOFF', format='1E', 
						array=numpy.zeros((nSource,), dtype=numpy.float32))
		# Mean equinox and epoch
		c12 = pyfits.Column(name='EQUINOX', format='A8',
						array=numpy.array(('J2000',)).repeat(nSource))
		c13 = pyfits.Column(name='EPOCH', format='1D', 
						array=numpy.zeros((nSource,), dtype=numpy.float64) + 2000.0)
		# Appearrent right ascension in degrees
		c14 = pyfits.Column(name='RAAPP', format='1D', 
						array=numpy.array(raList))
		# Apparent declination in degrees
		c15 = pyfits.Column(name='DECAPP', format='1D', 
						array=numpy.array(decList))
		# Right ascension at mean equinox in degrees
		c16 = pyfits.Column(name='RAEPO', format='1D', 
						array=numpy.array(raPoList))
		# Declination at mean equinox in degrees
		c17 = pyfits.Column(name='DECEPO', format='1D', 
						array=numpy.array(decPoList))
		# Systemic velocity in m/s
		c18 = pyfits.Column(name='SYSVEL', format='1D', 
						array=numpy.zeros((nSource,), dtype=numpy.float64))
		# Velocity type
		c19 = pyfits.Column(name='VELTYP', format='A8', 
						array=numpy.array(('GEOCENTR',)).repeat(nSource))
		# Velocity definition
		c20 = pyfits.Column(name='VELDEF', format='A8', 
						array=numpy.array(('OPTICAL',)).repeat(nSource))
		# Line rest frequency in Hz
		c21 = pyfits.Column(name='RESTFREQ', format='1D', 
						array=(numpy.zeros((nSource,), dtype=numpy.float64) + self.refVal))
		# Proper motion in RA in degrees/day
		c22 = pyfits.Column(name='PMRA', format='1D', 
						array=numpy.zeros((nSource,), dtype=numpy.float64))
		# Proper motion in Dec in degrees/day
		c23 = pyfits.Column(name='PMDEC', format='1D', 
						array=numpy.zeros((nSource,), dtype=numpy.float64))
		# Parallax of source in arc sec.
		c24 = pyfits.Column(name='PARALLAX', format='1E', 
						array=numpy.zeros((nSource,), dtype=numpy.float32))

		# Define the collection of columns
		colDefs = pyfits.ColDefs([c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, 
							c11, c12, c13, c14, c15, c16, c17, c18, c19, c20, 
							c21, c22, c23, c24])

		# Create the Source table and update its header
		sr = pyfits.new_table(colDefs)
		self.__addCommonKeywords(sr.header, 'SOURCE', 1)
		
		sr.name = 'SOURCE'
		self.FITS.append(sr)
		self.FITS.flush()

	def __writeData(self):
		"""Define the UV_Data table (group 3, table 1)."""

		hrz = astro.hrz_posn(0, 90)
		(arrPos, ag) = self.readArrayGeometry()
		(mapper, inverseMapper) = self.readArrayMapper()
		ids = ag.keys()

		mList = []
		uList = []
		vList = []
		wList = []
		timeList = []
		dateList = []
		intTimeList = []
		blineList = []
		nameList = []
		sourceList = []
		sourceID = 1
		for dataSet in self.data:
			utc = astro.taimjd_to_utcjd(dataSet.obsTime)
			date = astro.get_date(utc)
			date.hours = 0
			date.minutes = 0
			date.seconds = 0
			utc0 = date.to_jd()
			equ = hrz.to_equ(arrPos, utc)
			
			# format 'source' name based on local sidereal time
			raHms = astro.deg_to_hms(equ.ra)
			(tsecs, secs) = math.modf(raHms.seconds)
			name = "ZA%02d%02d%02d%01d" % (raHms.hours, raHms.minutes, int(secs), 
				int(tsecs * 10.0))
			nameList.append(name)
			
			# Compute the uvw coordinates of all baselines
			if inverseMapper is not None:
				standIDs = []
				for stand in self.FITS['ARRAY_GEOMETRY'].data.field('NOSTA'):
					standIDs.append(inverseMapper[stand])
				standIDs = numpy.array(standIDs)
			else:
				standIDs = self.FITS['ARRAY_GEOMETRY'].data.field('NOSTA')
			uvwBaselines = numpy.array([(i << 16) | j for i,j in uvUtils.getBaselines(standIDs)])
			uvwCoords = uvUtils.computeUVW(standIDs, HA=0.0, dec=equ.dec, freq=self.refVal)
			uvwCoords *= 1.0 / self.refVal

			# Loop over the data store in the dataDict and extract each baseline
			baselines = list(dataSet.dataDict.keys())
			baselines.sort()
			for baseline in baselines: 
		
				# validate baseline antenna ID's
				stand1 = (baseline >> 16) & 65535
				stand2 = baseline & 65535
				if mapper is not None:
					stand1 = mapper[stand1]
					stand2 = mapper[stand2]
				# Recontruct the baseline in FITS IDI format
				baselineMapped = (stand1 << 8) | stand2

				if (stand1 not in ids) or (stand2 not in ids):
					raise ValueError("baseline 0x%04x (%i to %i) contains unknown antenna IDs" % (baselineMapped, stand1, stand2))
			
				# validate data matrix shape
				visData = dataSet.dataDict[baseline]
				if (len(visData) != self.nChan):
					raise ValueError("data for baseline 0x%04x is the wrong size %s (!= %s)" % (baselineMapped, len(visData), self.nChan))
			
				# Calculate the UVW coordinates for antenna pair.  Catch auto-
				# correlations and set their UVW's to zeros.
				try:
					which = (numpy.where( uvwBaselines == baseline ))[0][0]
					uvw = numpy.squeeze(uvwCoords[which,:,0])
				except IndexError:
					uvw = numpy.zeros((3,))

				# Load the data into a matrix that splits the real and imaginary parts out
				matrix = numpy.zeros((2*self.nChan,), dtype=numpy.float32)
				matrix[0::2] = visData.real
				matrix[1::2] = visData.imag
			
				blineList.append(baselineMapped)
				mList.append(matrix.ravel())
				uList.append(uvw[0])
				vList.append(uvw[1])
				wList.append(uvw[2])
				dateList.append(utc0)
				timeList.append(utc - utc0) 
				intTimeList.append(dataSet.intTime)
				sourceList.append(sourceID)
			sourceID += 1
		nBaseline = len(blineList)
		nSource = len(nameList)

		# Visibility Data
		c1 = pyfits.Column(name='FLUX', format='%iE' % (2*self.nChan), unit='UNCALIB', 
						array=numpy.array(mList, dtype=numpy.float32))
		# Baseline number (first*256+second)
		c2 = pyfits.Column(name='BASELINE', format='1J', 
						array=numpy.array(blineList))
		# Julian date at 0h
		c3 = pyfits.Column(name='DATE', format='1D', unit='DAYS',
						array = numpy.array(dateList))
		# Time elapsed since 0h
		c4 = pyfits.Column(name='TIME', format='1D', unit = 'DAYS', 
						array = numpy.array(timeList))
		# Inegration time (seconds)
		c5 = pyfits.Column(name='INTTIM', format='1D', unit='SECONDS', 
						array=numpy.array(intTimeList, dtype=numpy.float32))
		# U coordinate (light seconds)
		c6 = pyfits.Column(name='UU', format='1E', unit='SECONDS', 
						array=numpy.array(uList, dtype=numpy.float32))
		# V coordinate (light seconds)
		c7 = pyfits.Column(name='VV', format='1E', 
						array=numpy.array(vList, dtype=numpy.float32))
		# W coordinate (light seconds)
		c8 = pyfits.Column(name='WW', format='1E', 
						array=numpy.array(wList, dtype=numpy.float32))
		# Source ID number
		c9 = pyfits.Column(name='SOURCE', format='1J', 
						array=numpy.array(sourceList))
		# Frequency setup number
		c10 = pyfits.Column(name='FREQID', format='1J', 
						array=(numpy.zeros((nBaseline,), dtype=numpy.int32) + self.freq[0].id))
		# Filter number
		c11 = pyfits.Column(name='FILTER', format='1J', 
						array=numpy.zeros((nBaseline,), dtype=numpy.int32))
		# Gate ID number
		c12 = pyfits.Column(name='GATEID', format='1J', 
						array=numpy.zeros((nBaseline,), dtype=numpy.int32))
		# Weights
		c13 = pyfits.Column(name='WEIGHT', format='%iE' % self.nChan, 
						array=numpy.ones((nBaseline, self.nChan), dtype=numpy.float32))
		
		colDefs = pyfits.ColDefs([c6, c7, c8, c3, c4, c2, c11, c9, c10, c5, 
						c13, c12, c1])

		# Create the UV Data table and update its header
		uv = pyfits.new_table(colDefs)
		self.__addCommonKeywords(uv.header, 'UV_DATA', 1)

		uv.header.update('NMATRIX', 1, 'number of UV data matricies')
		uv.header.update('MAXIS', 6, 'number of UV data matrix axes')
		uv.header.update('TMATX13', True, 'axis 13 contains UV matrix')
		
		uv.header.update('MAXIS1', 2, 'number of pixels in COMPLEX axis')
		uv.header.update('CTYPE1', 'COMPLEX', 'axis 1 is COMPLEX axis')
		uv.header.update('CDELT1', 1.0)
		uv.header.update('CRPIX1', 1.0)
		uv.header.update('CRVAL1', 1.0)
		
		uv.header.update('MAXIS2', 1, 'number of pixels in STOKES axis')
		uv.header.update('CTYPE2', 'STOKES', 'axis 2 is STOKES axis (polarization)')
		uv.header.update('CDELT2', -1.0)
		uv.header.update('CRPIX2', 1.0)
		uv.header.update('CRVAL2', float(self.stokes[0]))
		
		uv.header.update('MAXIS3', self.nChan, 'number of pixels in FREQ axis')
		uv.header.update('CTYPE3', 'FREQ', 'axis 3 is FREQ axis (frequency)')
		uv.header.update('CDELT3', self.freq[0].chWidth)
		uv.header.update('CRPIX3', self.refPix)
		uv.header.update('CRVAL3', self.refVal)
		
		uv.header.update('MAXIS4', 1, 'number of pixels in BAND (IF) axis')
		uv.header.update('CTYPE4', 'BAND', 'axis 4 is BAND axis')
		uv.header.update('CDELT4', 1.0)
		uv.header.update('CRPIX4', 1.0)
		uv.header.update('CRVAL4', 1.0)
		
		uv.header.update('MAXIS5', 1, 'number of pixels in RA axis')
		uv.header.update('CTYPE5', 'RA', 'axis 5 is RA axis (position of phase center)')
		uv.header.update('CDELT5', 0.0)
		uv.header.update('CRPIX5', 1.0)
		uv.header.update('CRVAL5', 0.0)
		
		uv.header.update('MAXIS6', 1, 'number of pixels in DEC axis')
		uv.header.update('CTYPE6', 'DEC', 'axis 6 is DEC axis (position of phase center)')
		uv.header.update('CDELT6', 0.0)
		uv.header.update('CRPIX6', 1.0)
		uv.header.update('CRVAL6', 0.0)
		
		uv.header.update('TELESCOP', 'LWA')
		uv.header.update('OBSERVER', 'ZASKY')
		uv.header.update('SORT', 'TB', 'data is sorted in [time,baseline] order')
		
		uv.header.update('VISSCALE', 1.0, 'UV data scale factor')
		
		uv.name = 'UV_DATA'
		self.FITS.append(uv)
		self.FITS.flush()

	def __writeMapper(self):
		"""Write a fits table that contains information about mapping stations 
		numbers to actual antenna numbers.  This information can be backed out of
		the names, but this makes the extraction more programatic."""

		c1 = pyfits.Column(name='ANNAME', format='A8', 
						array=numpy.array([ant.getName() for ant in self.array[0]['ants']]))
		c2 = pyfits.Column(name='NOSTA', format='1J', 
						array=numpy.array([self.array[0]['mapper'][ant.id] for ant in self.array[0]['ants']]))
		c3 = pyfits.Column(name='NOACT', format='1J', 
						array=numpy.array([ant.id for ant in self.array[0]['ants']]))

		colDefs = pyfits.ColDefs([c1, c2, c3])

		# Create the ID mapping table and update its header
		nsm = pyfits.new_table(colDefs)
		self.__addCommonKeywords(nsm.header, 'NOSTA_MAPPER', 1)

		nsm.name = 'NOSTA_MAPPER'
		self.FITS.append(nsm)
		self.FITS.flush()

	def readArrayGeometry(self):
		"""Return a tuple with the array geodetic position and the local 
		positions for all antennas defined in the ARRAY_GEOMETRY table."""

		try:
			ag = self.FITS['ARRAY_GEOMETRY']
		except IndexError:
			raise RuntimeError("File does not have an 'ARRAY_GEOMTRY' table.")

		# Array position
		arrayGeo = astro.rect_posn(ag.header['ARRAYX'], ag.header['ARRAYY'], ag.header['ARRAYZ'])
		arrayGeo = astro.get_geo_from_rect(arrayGeo)

		# Antenna positions
		antennaGeo = {}
		antenna = ag.data
		antennaID = antenna.field('NOSTA')
		antennaPos = iter(antenna.field('STABXYZ'))
		for id in antennaID:
			antennaGeo[id] = antennaPos.next()

		# Return
		return (arrayGeo, antennaGeo)

	def readArrayMapper(self):
		"""Return a tuple with the array NOSTA mapper and inverse mapper (both
		dictionaries.  If the stand IDs have not been mapped, return None for
		both."""

		try:
			nsm = self.FITS['NOSTA_MAPPER']
		except IndexError:
			return (None, None)

		# Build the mapper and inverseMapper
		mapper = {}
		inverseMapper = {}
		nosta = nsm.data.field('NOSTA')
		noact = nsm.data.field('NOACT')
		for idMapped, idActual in zip(nosta, noact):
			mapper[idActual] = idMapped
			inverseMapper[idMapped] = idActual

		# Return
		return (mapper, inverseMapper)



# -*- coding: utf-8 -*-

"""Unit test for lsl.reader modules"""

import os
import unittest

from lsl.common.paths import dataBuild as dataPath
from lsl.reader import s60
from lsl.reader import tbw
from lsl.reader import tbn
from lsl.reader import drx
from lsl.reader import errors


__revision__ = "$Revision:1 $"
__version__  = "0.1"
__author__    = "Jayce Dowell"


s60File = os.path.join(dataPath, 'tests', 's60-test.dat')
tbwFile = os.path.join(dataPath, 'tests', 'tbw-test.dat')
tbnFile = os.path.join(dataPath, 'tests', 'tbn-test.dat')


class reader_tests(unittest.TestCase):
	"""A unittest.TestCase collection of unit tests for the lsl.reader
	modules."""

	### S60 ###

	def test_s60_read(self):
		"""Test reading in a frame from a S60 file."""

		fh = open(s60File, 'rb')
		# First frame makes it in with the correct number of elements
		frame1 = s60.readFrame(fh)
		self.assertEqual(frame1.shape[0], 734)

		# Next try a chunk that is equivalent to 3 frames
		chunk1 = s60.readChunk(fh, Chunk=2202)
		self.assertEqual(chunk1.shape[0], 2202)
		fh.close()

	def test_s60_errors(self):
		"""Test reading in all frames from a truncated S60 file."""

		fh = open(s60File, 'rb')
		# Frames 1 through 6
		for i in range(1,7):
			frame = s60.readFrame(fh)

		# Last frame should be an error (errors.numpyError)
		self.assertRaises(errors.numpyError, s60.readFrame, fh)
		fh.close()

	### TBW ###

	def test_tbw_read(self):
		"""Test reading in a frame from a TBW file."""

		fh = open(tbwFile, 'rb')
		# First frame is really TBW and stores the correct stand ID
		frame1 = tbw.readFrame(fh)
		self.assertTrue(frame1.header.isTBW())
		self.assertEqual(frame1.parseID(), 2)
		# Second frame
		frame2 = tbw.readFrame(fh)
		self.assertTrue(frame2.header.isTBW())
		self.assertEqual(frame2.parseID(), 1)
		fh.close()

	def test_tbw_bits(self):
		"""Test getting the data bits from a TBW file."""

		fh = open(tbwFile, 'rb')
		# File contains 12-bit data, two ways
		self.assertEqual(tbw.getDataBits(fh), 12)
		frame1 = tbw.readFrame(fh)
		self.assertEqual(frame1.getDataBits(), 12)
		fh.close()

	def test_tbw_errors(self):
		"""Test reading errors."""

		fh = open(tbwFile, 'rb')
		# Frames 1 through 8
		for i in range(1,9):
			frame = tbw.readFrame(fh)

		# Last frame should be an error (errors.numpyError)
		self.assertRaises(errors.numpyError, tbw.readFrame, fh)
		fh.close()
		
		# If we offset in the file by 1 byte, we should be a 
		# sync error (errors.syncError).
		fh = open(tbwFile, 'rb')
		fh.seek(1)
		self.assertRaises(errors.syncError, tbw.readFrame, fh)
		fh.close()

	### TBN ###

	def test_tbn_read(self):
		"""Test reading in a frame from a TBN file."""

		fh = open(tbnFile, 'rb')
		# First frame is really TBW and stores the correct stand ID
		frame1 = tbn.readFrame(fh)
		stand, pol = frame1.parseID()
		self.assertEqual(stand, 1)
		self.assertEqual(pol, 0)
		# Second frame
		frame2 = tbn.readFrame(fh)
		stand, pol = frame2.parseID()
		self.assertEqual(stand, 1)
		self.assertEqual(pol, 1)
		fh.close()

	def test_tbn_errors(self):
		"""Test reading in all frames from a truncated TBN file."""

		fh = open(tbnFile, 'rb')
		# Frames 1 through 29
		for i in range(1,30):
			frame = tbn.readFrame(fh)

		# Last frame should be an error (errors.numpyError)
		self.assertRaises(errors.numpyError, tbn.readFrame, fh)
		fh.close()
		
		# If we offset in the file by 1 byte, we should be a 
		# sync error (errors.syncError).
		fh = open(tbnFile, 'rb')
		fh.seek(1)
		self.assertRaises(errors.syncError, tbn.readFrame, fh)
		fh.close()

	def test_tbn_block(self):
		"""Test finding out how many stands are in a file."""

		fh = open(tbnFile, 'rb')
		nx, ny = tbn.getFramesPerObs(fh)
		self.assertEqual(nx, 10)
		self.assertEqual(ny, 10)
		fh.close()

	def test_tbn_rate(self):
		"""Test finding out the sample rate of a TBN file."""

		fh = open(tbnFile, 'rb')
		rate = tbn.getSampleRate(fh)
		self.assertEqual(rate, 100000)
		code = tbn.getSampleRate(fh, FilterCode=True)
		self.assertEqual(code, 7)
		fh.close()

	### TBW/TBN Mix-up ###

	def test_tbw_tbn_catch(self):
		"""Test that tbw will not read tbn files and vice versa."""

		fh = open(tbnFile, 'rb')
		frame1 = tbw.readFrame(fh)
		self.assertFalse(frame1.header.isTBW())
		fh.close()

		fh = open(tbwFile, 'rb')
		frame1 = tbn.readFrame(fh)
		self.assertFalse(frame1.header.isTBN())
		fh.close()
		
	### DRX ###

class reader_test_suite(unittest.TestSuite):
	"""A unittest.TestSuite class which contains all of the lsl.reader units 
	tests."""
	
	def __init__(self):
		unittest.TestSuite.__init__(self)
		
		loader = unittest.TestLoader()
		self.addTests(loader.loadTestsFromTestCase(reader_tests)) 


if __name__ == '__main__':
	unittest.main()



# -*- coding: utf-8 -*-

"""Python module to reading in data from both 12-bit and 4-bit TBW files.  
This module defines the following classes for storing the TBW data found in
a file:

Frame
  object that contains all data associated with a particular TBW frame.  The 
  primary consituents of each frame are:
    * FrameHeader - the TBW frame header object and
    * FrameData   - the TBW frame data object.  
  Combined, these two objects contain all of the information found in the 
  original TBW frame.

The functions defined in this module fall into two class:
  1. convert a frame in a file to a Frame object and
  2. describe the format of the data in the file.

For reading in data, use the readFrame function.  It takes a python file-
handle as an input and returns a fully-filled Frame object.  readFrame 
is designed to work with both 4-bit and 12-bit observations.

For describing the format of data in the file, two function are provided:

getDataBits
  read in the first frame of an open file handle and return whether or not 
  the data is 12 or 4-bit

getFramesPerObs
  read in the first several frames to see how many stands are found in the 
  data.
  .. note::

	This function is a little flaky on TBW data sets that have less 
	than a full complement or 12M (36M) samples.
"""

import os
import sys
import time
import numpy
import array
import struct
import pyfits

from  lsl.common import dp as dp_common
from errors import *

__version__ = '0.5'
__revision__ = '$ Revision: 18 $'
__all__ = ['FrameHeader', 'FrameData', 'Frame', 'readFrame', 'FrameSize', 'getDataBits', 'getFramesPerObs', '__version__', '__revision__', '__all__']

FrameSize = 1224


class FrameHeader(object):
	"""Class that stores the information found in the header of a TBW 
	frame.  All three fields listed in the DP IDC version H are stored as 
	well as the original binary header data."""

	def __init__(self, frameCount=None, secondsCount=None, tbwID=None,  raw=None):
		self.frameCount = frameCount
		self.secondsCount = secondsCount
		self.tbwID = tbwID
		self.raw = raw

	def isTBW(self):
		"""Function to check if the data is really TBW and not TBN by examining
		the TBW ID field.  Returns True if the data is TBW, false otherwise."""

		mode = (self.tbwID>>15)&1
		if mode == 1:
			return True
		else:
			return False

	def parseID(self):
		"""Function to parse the TBW ID field and return the stand number."""

		# Why &1023?  Well, from DP ICH revision H, it seems that the stand count 
		# only goes up 260.  So, channel numbers should range up to 520, which can
		# be represented as 10 bits or 1023.
		stand = self.tbwID&1023

		return stand

	def getDataBits(self):
		"""Function to parse the TBW ID field and return the size of number of 
		bits that comprise the data.  12 is returned for 12-bit data, and 4 
		for 4-bit data."""

		bits = (self.tbwID>>14)&1
		if bits == 0:
			dataBits = 12
		else:
			dataBits = 4

		return dataBits


class FrameData(object):
	"""Class that stores the information found in the data section of a TBW
	frame.  Both fields listed in the DP IDC version H are stored."""

	def __init__(self, timeTag=None, samples=400, xy=None):
		self.timeTag = timeTag
		self.xy = xy

	def getTime(self):
		"""Function to convert the time tag from samples since station 
		midnight to seconds since station midnight.  This function needs to 
		dp_common module in order to work."""

		seconds = self.timeTag / dp_common.fS
		
		return seconds


class Frame(object):
	"""Class that stores the information contained within a single TBW 
	frame.  It's properties are FrameHeader and FrameData objects."""

	def __init__(self, header=FrameHeader(), data=FrameData()):
		self.header = header
		self.data = data
		self.valid = True

	def parseID(self):
		"""Convenience wrapper for the Frame.FrameHeader.parseID 
		function."""
		
		return self.header.parseID()

	def getDataBits(self):
		"""Convenience wrapper for the Frame.FrameHeader.getDataBits 
		function."""
		
		return self.header.getDataBits()

	def getTime(self):
		"""Convenience wrapper for the Frame.FrameData.getTime function."""
		
		return self.data.getTime()
			
	def __add__(self, y):
		"""Add the data sections of two frames together or add a number 
		to every element in the data section."""
		
		try:
			return self.data.xy + y.self.data.xy
		except AttributeError:
			return self.data.xy + y
			
	def __iadd__(self, y):
		"""In-place add the data sections of two frames together or add 
		a number to every element in the data section."""
		
		try:
			self.data.xy += y.self.data.xy
		except AttributeError:
			self.data.xy += y
		
	def __mul__(self, y):
		"""Multiple the data sections of two frames together or multiply 
		a number to every element in the data section."""
		
		try:
			return self.data.xy * y.self.data.xy
		except AttributeError:
			return self.data.xy * y
			
	def __imul__(self, y):
		"""In-place multiple the data sections of two frames together or 
		multiply a number to every element in the data section."""
		
		try:
			self.data.xy *= y.self.data.xy
		except AttributeError:
			self.data.xy *= y
			
	def __eq__(self, y):
		"""Check if the time tags of two frames are equal or if the time
		tag is equal to a particular value."""
		
		tX = self.data.timeTag
		try:
			tY = self.data.timeTag
		except AttributeError:
			tY = y
		
		if tX == tY:
			return True
		else:
			return False
			
	def __ne__(self, y):
		"""Check if the time tags of two frames are not equal or if the time
		tag is not equal to a particular value."""
		
		tX = self.data.timeTag
		try:
			tY = self.data.timeTag
		except AttributeError:
			tY = y
		
		if tX != tY:
			return True
		else:
			return False
			
	def __gt__(self, y):
		"""Check if the time tag of the first frame is greater than that of a
		second frame or if the time tag is greater than a particular value."""
		
		tX = self.data.timeTag
		try:
			tY = self.data.timeTag
		except AttributeError:
			tY = y
		
		if tX > tY:
			return True
		else:
			return False
			
	def __ge__(self, y):
		"""Check if the time tag of the first frame is greater than or equal to 
		that of a second frame or if the time tag is greater than a particular 
		value."""
		
		tX = self.data.timeTag
		try:
			tY = self.data.timeTag
		except AttributeError:
			tY = y
		
		if tX >= tY:
			return True
		else:
			return False
			
	def __lt__(self, y):
		"""Check if the time tag of the first frame is less than that of a
		second frame or if the time tag is greater than a particular value."""
		
		tX = self.data.timeTag
		try:
			tY = self.data.timeTag
		except AttributeError:
			tY = y
		
		if tX < tY:
			return True
		else:
			return False
			
	def __le__(self, y):
		"""Check if the time tag of the first frame is less than or equal to 
		that of a second frame or if the time tag is greater than a particular 
		value."""
		
		tX = self.data.timeTag
		try:
			tY = self.data.timeTag
		except AttributeError:
			tY = y
		
		if tX <= tY:
			return True
		else:
			return False
			
	def __cmp__(self, y):
		"""Compare two frames based on the time tags.  This is helpful for 
		sorting things."""
		
		tX = self.data.timeTag
		tY = y.data.timeTag
		if tY > tX:
			return -1
		elif tX > tY:
			return 1
		else:
			return 0


def __readHeader(filehandle, Verbose=False):
	"""Private function to read in a TBW header.  Returns a FrameHeader object."""

	rawHeader = ''
	try:
		s = filehandle.read(4)
		rawHeader = rawHeader + s
		sync4, sync3, sync2, sync1 = struct.unpack(">BBBB", s)
		s = filehandle.read(4)
		rawHeader = rawHeader + s
		m5cID, frameCount3, frameCount2, frameCount1 = struct.unpack(">BBBB", s)
		frameCount = (long(frameCount3)<<16) | (long(frameCount2)<<8) | long(frameCount1)
		s = filehandle.read(4)
		rawHeader = rawHeader + s
		secondsCount = struct.unpack(">L", s)
		s = filehandle.read(4)
		rawHeader = rawHeader + s
		tbwID, junk = struct.unpack(">HH", s)
	except IOError:
		raise eofError()
	except struct.error:
		raise eofError()

	if sync1 != 92 or sync2 != 222 or sync3 != 192 or sync4 != 222:
		raise syncError(sync1=sync1, sync2=sync2, sync3=sync3, sync4=sync4)

	newHeader = FrameHeader()
	newHeader.frameCount = frameCount
	newHeader.secondsCount = secondsCount[0]
	newHeader.tbwID = tbwID

	if Verbose:
		stand = newHeader.parseID()
		print "Header: ", tbwID, secondsCount, junk
		print "  Stand: ", stand

	return newHeader


def __readData12(filehandle):
	"""Private function to read in a TBW frame data section and unpack that data
	when is the 12-bit.  Returns a FrameData object."""

	try:
		s = filehandle.read(8)
		timeTag = struct.unpack(">Q", s)
	except IOError:
		raise eofError()
	except struct.error:
		raise eofError()

	rawData = numpy.fromfile(filehandle, dtype=numpy.uint8, count=1200)
	rawData = rawData.astype(numpy.uint16)
	data = numpy.zeros((2,400), dtype=numpy.int16)
	if rawData.shape[0] < 3*data.shape[1]:
		raise numpyError()

	data[0,:] = (rawData[0::3]<<4) | ((rawData[1::3]>>4)&15)
	data[1,:] = ((rawData[1::3]&15)<<8) | rawData[2::3]
	# The data are signed, so apply the two-complement rule to generate 
	# the negative values
	negativeValues = numpy.where( data >= 2048 )
	data[negativeValues] -= 4096

	newData = FrameData()
	newData.timeTag = long(timeTag[0])
	newData.xy = data

	return newData
	
	
def __readData4(filehandle):
	"""Private function to read in a TBW frame data section and unpack that data
	when is the 4-bit.  Returns a FrameData object."""

	try:
		s = filehandle.read(8)
		timeTag = struct.unpack(">Q", s)
	except IOError:
		raise eofError()
	except struct.error:
		raise eofError()
	
	rawData = numpy.fromfile(filehandle, dtype=numpy.uint8, count=1200)
	data = numpy.zeros((2,1200), dtype=numpy.int8)
	if rawData.shape[0] < data.shape[1]:
		raise numpyError()

	data[0,:] = (rawData>>4)&15
	data[1,:] = rawData&15
	# The data are signed, so apply the two-complement rule to generate 
	# the negative values
	negativeValues = numpy.where( data >= 8 )
	data[negativeValues] -= 16
	
	newData = FrameData()
	newData.timeTag = long(timeTag[0])
	newData.xy = data
	
	return newData


def readFrame(filehandle, Verbose=False):
	"""Function to read in a single TBW frame (header+data) and store the 
	contents as a Frame object.  This function wraps readerHeader and 
	readData[(12)|4]."""

	try:
		hdr = __readHeader(filehandle, Verbose=Verbose)
	except syncError, err:
		# Why?  If we run into a sync error here, then the following frame is invalid.  
		# Thus, we need to skip over this frame be advancing the file pointer 8+1200 B 
		currPos = filehandle.tell()
		frameEnd = currPos + FrameSize - 16
		filehandle.seek(frameEnd)
		raise err

	if hdr.getDataBits() == 12:
		dat = __readData12(filehandle)
	else:
		dat = __readData4(filehandle)
	
	newFrame = Frame()
	newFrame.header = hdr
	newFrame.data = dat

	return newFrame


def getDataBits(filehandle):
	"""Find out the number of data bits used in the file be reading in the 
	first frame."""

	# Save the current position in the file so we can return to that point
	fhStart = filehandle.tell()
	
	# Go back to the beginning...
	filehandle.seek(0)

	cFrame = readFrame(filehandle)

	dataBits = cFrame.getDataBits()

	# Return to the place in the file where we started
	filehandle.seek(fhStart)

	return dataBits


def getFramesPerObs(filehandle):
	"""Find out how many frames are present per observation by examining 
	the first frames for what would be 256 stands.  This is done by reading
	two frames and then skipping the next 30,000."""

	# Save the current position in the file so we can return to that point
	fhStart = filehandle.tell()
	
	# Go back to the beginning...
	filehandle.seek(0)

	idCodes = []
	for i in range(3*256):
		currentPosition = filehandle.tell()
		try:
			cFrame1 = readFrame(filehandle)
			cFrame2 = readFrame(filehandle)
		except eofError:
			break
		except syncError:
			continue
		except numpyError:
			break

		cID = cFrame1.parseID()
		if cID not in idCodes:
			idCodes.append(cID)
		cID = cFrame2.parseID()
		if cID not in idCodes:
			idCodes.append(cID)

		# Junk 30,000 frames since that is how many frames there are per stand
		filehandle.seek(currentPosition+30000*FrameSize)

	# Return to the place in the file where we started
	filehandle.seek(fhStart)
	
	# Get the length of the stand list and return
	return len(idCodes)



# -*- coding: utf-8 -*-

"""Unit test for the lsl.writer.fitsidi modules."""

import os
import time
import unittest
import tempfile
import numpy
import pyfits

from lsl.common import stations as lwa_common
from lsl.correlator import uvUtils
from lsl.writer import fitsidi


__revision__ = "$Revision:1 $"
__version__  = "0.1"
__author__    = "Jayce Dowell"


class fitsidi_tests(unittest.TestCase):
	"""A unittest.TestCase collection of unit tests for the lsl.writer.fitsidi
	module."""

	def __initData(self):
		"""Private function to generate a random set of data for writing a FITS
		IDI file.  The data is returned as a dictionary with keys:
		* freq - frequency array in Hz
		* site - lwa.common.stations object
		* stands - array of stand numbers
		* bl - list of baseline pairs in real stand numbers
		* vis - array of visibility data in baseline x freq format
		"""

		# Frequency range
		freq = numpy.arange(0,512)*20e6/512 + 40e6
		# Site and stands
		site = lwa_common.lwa1()
		stands = site.getStands('2010/11/25 00:00:00')
		# Set baselines and data
		blList = uvUtils.getBaselines(stands, IncludeAuto=True, Indicies=False)
		visData = numpy.random.rand(len(blList), len(freq))
		visData = visData.astype(numpy.complex64)

		return {'freq': freq, 'site': site, 'stands': stands, 'bl': blList, 'vis': visData}

	def test_write_tables(self):
		"""Test if the FITS IDI writer writes all of the tables."""

		testTime = time.time()
		testPath = tempfile.mkdtemp(prefix='fitsidi-', suffix='.tmp')
		testFile = os.path.join(testPath, 'idi-test.fits')
		
		# Get some data
		data = self.__initData()
		
		# Start the file
		fits = fitsidi.IDI(testFile, refTime=testTime)
		fits.setStokes(['xx'])
		fits.setFrequency(data['freq'])
		fits.setGeometry(data['site'], data['stands'])
		fits.addDataSet(testTime, 6.0, data['bl'], data['vis'])
		fits.write()

		# Open the file and examine
		hdulist = pyfits.open(testFile)
		# Check that all of the extensions are there
		extNames = [hdu.name for hdu in hdulist]
		for ext in ['ARRAY_GEOMETRY', 'FREQUENCY', 'ANTENNA', 'BANDPASS', 'SOURCE', 'UV_DATA']:
			self.assertTrue(ext in extNames)

		hdulist.close()
		os.unlink(testFile)
		os.rmdir(testPath)

	def test_array_geometry(self):
		"""Test the ARRAY_GEOMETRY table."""

		testTime = time.time()
		testPath = tempfile.mkdtemp(prefix='fitsidi-', suffix='.tmp')
		testFile = os.path.join(testPath, 'idi-test.fits')
		
		# Get some data
		data = self.__initData()
		
		# Start the file
		fits = fitsidi.IDI(testFile, refTime=testTime)
		fits.setStokes(['xx'])
		fits.setFrequency(data['freq'])
		fits.setGeometry(data['site'], data['stands'])
		fits.addDataSet(testTime, 6.0, data['bl'], data['vis'])
		fits.write()

		# Open the file and examine
		hdulist = pyfits.open(testFile)
		ag = hdulist['ARRAY_GEOMETRY'].data
		# Correct number of stands
		self.assertEqual(len(data['stands']), len(ag.field('NOSTA')))

		# Correct stand names
		names = ['LWA%03i' % stand for stand in data['stands']]
		for name, anname in zip(names, ag.field('ANNAME')):
			self.assertEqual(name, anname)

		hdulist.close()
		os.unlink(testFile)
		os.rmdir(testPath)

	def test_frequency(self):
		"""Test the FREQUENCY table."""

		testTime = time.time()
		testPath = tempfile.mkdtemp(prefix='fitsidi-', suffix='.tmp')
		testFile = os.path.join(testPath, 'idi-test.fits')
		
		# Get some data
		data = self.__initData()
		
		# Start the file
		fits = fitsidi.IDI(testFile, refTime=testTime)
		fits.setStokes(['xx'])
		fits.setFrequency(data['freq'])
		fits.setGeometry(data['site'], data['stands'])
		fits.addDataSet(testTime, 6.0, data['bl'], data['vis'])
		fits.write()

		# Open the file and examine
		hdulist = pyfits.open(testFile)
		fq = hdulist['FREQUENCY'].data
		# Correct number of FREQIDs
		self.assertEqual(len(fq.field('FREQID')), 1)

		# Correct channel width
		self.assertAlmostEqual(fq.field('CH_WIDTH')[0], numpy.abs(data['freq'][1]-data['freq'][0]), 4)

		# Correct bandwidth
		self.assertAlmostEqual(fq.field('TOTAL_BANDWIDTH')[0], numpy.abs(data['freq'][-1]-data['freq'][0]).astype(numpy.float32), 4)

		# Correct sideband
		self.assertEqual(fq.field('SIDEBAND')[0], 1)

		hdulist.close()
		os.unlink(testFile)
		os.rmdir(testPath)

	def test_antenna(self):
		"""Test the ANTENNA table."""

		testTime = time.time()
		testPath = tempfile.mkdtemp(prefix='fitsidi-', suffix='.tmp')
		testFile = os.path.join(testPath, 'idi-test.fits')
		
		# Get some data
		data = self.__initData()
		
		# Start the file
		fits = fitsidi.IDI(testFile, refTime=testTime)
		fits.setStokes(['xx'])
		fits.setFrequency(data['freq'])
		fits.setGeometry(data['site'], data['stands'])
		fits.addDataSet(testTime, 6.0, data['bl'], data['vis'])
		fits.write()

		# Open the file and examine
		hdulist = pyfits.open(testFile)
		an = hdulist['ANTENNA'].data
		# Correct number of stands
		self.assertEqual(len(data['stands']), len(an.field('ANTENNA_NO')))

		# Correct FREQIDs
		for freqid in an.field('FREQID'):
			self.assertEqual(freqid, 1)

		hdulist.close()
		os.unlink(testFile)
		os.rmdir(testPath)

	def test_bandpass(self):
		"""Test the BANDPASS table."""

		testTime = time.time()
		testPath = tempfile.mkdtemp(prefix='fitsidi-', suffix='.tmp')
		testFile = os.path.join(testPath, 'idi-test.fits')
		
		# Get some data
		data = self.__initData()
		
		# Start the file
		fits = fitsidi.IDI(testFile, refTime=testTime)
		fits.setStokes(['xx'])
		fits.setFrequency(data['freq'])
		fits.setGeometry(data['site'], data['stands'])
		fits.addDataSet(testTime, 6.0, data['bl'], data['vis'])
		fits.write()

		# Open the file and examine
		hdulist = pyfits.open(testFile)
		bp = hdulist['BANDPASS'].data
		# Correct number of entries
		self.assertEqual(len(data['stands']), len(bp.field('ANTENNA_NO')))

		# Correct Source ID number
		for src in bp.field('SOURCE_ID'):
			self.assertEqual(src, 0)

		# Correct FREQIDs
		for freqid in bp.field('FREQID'):
			self.assertEqual(freqid, 1)

		hdulist.close()
		os.unlink(testFile)
		os.rmdir(testPath)

	def test_source(self):
		"""Test the SOURCE table."""

		testTime = time.time()
		testPath = tempfile.mkdtemp(prefix='fitsidi-', suffix='.tmp')
		testFile = os.path.join(testPath, 'idi-test.fits')
		
		# Get some data
		data = self.__initData()
		
		# Start the file
		fits = fitsidi.IDI(testFile, refTime=testTime)
		fits.setStokes(['xx'])
		fits.setFrequency(data['freq'])
		fits.setGeometry(data['site'], data['stands'])
		fits.addDataSet(testTime, 6.0, data['bl'], data['vis'])
		fits.write()

		# Open the file and examine
		hdulist = pyfits.open(testFile)
		so = hdulist['SOURCE'].data
		# Correct number of entries
		self.assertEqual(len(so.field('SOURCE_ID')), 1)

		# Correct Source ID number
		self.assertEqual(so.field('SOURCE_ID'), 1)

		hdulist.close()
		os.unlink(testFile)
		os.rmdir(testPath)

	def test_uvdata(self):
		"""Test the UV_DATA table."""

		testTime = time.time()
		testPath = tempfile.mkdtemp(prefix='fitsidi-', suffix='.tmp')
		testFile = os.path.join(testPath, 'idi-test.fits')
		
		# Get some data
		data = self.__initData()
		
		# Start the file
		fits = fitsidi.IDI(testFile, refTime=testTime)
		fits.setStokes(['xx'])
		fits.setFrequency(data['freq'])
		fits.setGeometry(data['site'], data['stands'])
		fits.addDataSet(testTime, 6.0, data['bl'], data['vis'])
		fits.write()

		# Open the file and examine
		hdulist = pyfits.open(testFile)
		uv = hdulist['UV_DATA'].data

		# Load the mapper
		try:
			mp = hdulist['NOSTA_MAPPER'].data
			nosta = mp.field('NOSTA')
			noact = mp.field('NOACT')
		except KeyError:
			ag = hdulist['ARRAY_GEOMETRY'].data
			nosta = ag.field('NOSTA')
			noact = ag.field('NOACT')
		mapper = {}
		for s,a in zip(nosta, noact):
			mapper[s] = a

		# Correct number of visibilities
		self.assertEqual(len(uv.field('FLUX')), data['vis'].shape[0])
		
		# Correct number of frequencies
		for vis in uv.field('FLUX'):
			self.assertEqual(len(vis), 2*len(data['freq']))

		# Correct values
		for bl, vis in zip(uv.field('BASELINE'), uv.field('FLUX')):
			# Convert mapped stands to real stands
			stand1 = mapper[(bl >> 8) & 255]
			stand2 = mapper[bl & 255]

			# Find out which visibility set in the random data corresponds to the 
			# current visibility
			i = 0
			for s1,s2 in data['bl']:
				if s1 == stand1 and s2 == stand2:
					break
				else:
					i = i + 1
			
			# Extract the data and run the comparison
			visData = numpy.zeros(len(data['freq']), dtype=numpy.complex64)
			visData.real = vis[0::2]
			visData.imag = vis[1::2]
			for vd, sd in zip(visData, data['vis'][i,:]):
				self.assertAlmostEqual(vd, sd, 8)
			i = i + 1
		
		hdulist.close()
		os.unlink(testFile)
		os.rmdir(testPath)

	def test_mapper(self):
		"""Test the NOSTA_MAPPER table."""

		testTime = time.time()
		testPath = tempfile.mkdtemp(prefix='fitsidi-', suffix='.tmp')
		testFile = os.path.join(testPath, 'idi-test.fits')
		
		# Get some data
		data = self.__initData()
		
		# Start the file
		fits = fitsidi.IDI(testFile, refTime=testTime)
		fits.setStokes(['xx'])
		fits.setFrequency(data['freq'])
		fits.setGeometry(data['site'], data['stands'])
		fits.addDataSet(testTime, 6.0, data['bl'], data['vis'])
		fits.write()

		# Open the file and examine
		hdulist = pyfits.open(testFile)
		extNames = [hdu.name for hdu in hdulist]
		if data['stands'].max() > 255:
			self.assertTrue('NOSTA_MAPPER' in extNames)

			# Make sure the mapper makes sense
			mp = hdulist['NOSTA_MAPPER'].data
			ag = hdulist['ARRAY_GEOMETRY'].data
			mNoSta = mp.field('NOSTA')
			aNoSta = ag.field('NOSTA')
			mNoAct = mp.field('NOACT')
			aAnNam = ag.field('ANNAME')
			for msta, mact, asta, anam in zip(mNoSta, mNoAct, aNoSta, aAnNam):
				self.assertEqual(msta, asta)
				self.assertEqual(mact, int(anam[3:]))

		hdulist.close()
		os.unlink(testFile)
		os.rmdir(testPath)


class fitsidi_test_suite(unittest.TestSuite):
	"""A unittest.TestSuite class which contains all of the lsl.reader units 
	tests."""
	
	def __init__(self):
		unittest.TestSuite.__init__(self)
		
		loader = unittest.TestLoader()
		self.addTests(loader.loadTestsFromTestCase(fitsidi_tests))


if __name__ == '__main__':
	unittest.main()


# -*- coding: utf-8 -*

"""This module stores various functions that are needed for computing UV 
coverage and time delays.  The functions in the module:
  * return the x, y, and z coordinates of a stand or array of stands
  * return the relative x, y, and z offsets between two stands
  * return the cable delays as a function of frequency for a stand
  * compute the u, v, and w coordinates of all baselines defined by an array 
    of stands
  * compute the track through the uv-plane of a collection of baselines as 
    the Earth rotates.
"""

import os
import csv
import math
import numpy

from lsl.common.paths import data as dataPath
from lsl.common.constants import *

__version__ = '0.3'
__revision__ = '$ Revision: 15 $'
__all__ = ['validateStand', 'getXYZ', 'getRelativeXYZ', 'PositionCache', 'cableDelay', 'CableCache', 'signalDelay', 'SignalCache', 'cableAttenuation', 'getBaselines', 'baseline2antenna', 'antenna2baseline', 'computeUVW', 'computeUVTrack', 'uvUtilsError', '__version__', '__revision__', '__all__']


class uvUtilsError(Exception):
	"""Base exception class for this module"""

	def __init__(self, strerror):
		self.strerror = strerror
	
	def __str__(self):
		return "%s" % self.strerror


def validateStand(stand, max=258):
	"""Make sure that the provided stand number is valid.  If it is not between
	1 and max (default of 258 for LWA-1), raise an error."""

	if stand < 1 or stand > max:
			raise uvUtilsError('Stand #%i is out of range (1-%i)' % (stand, max))

	return True


def _loadPositionData(filename='lwa1-positions.csv'):
	"""Private function to load in the stand location data CSV file.  Someday 
	this should be absorbed into this script so that there is not need to keep 
	yet another file floating around."""

	# lwa1-asbuilt.csv contains the stand (x,y,z) coordinates for all 257 stands +
	# the outlier (#258).  The data come from the "LWA-1 Antenna Position and 
	# Cable Data" memo, version 1.
	try:
		csvFH = open(os.path.join(dataPath, filename), 'r')
	except IOError, (errno, strerror):
		raise uvUtilsError("%s: stand position file '%s'" % (strerror, filename))
	csvData = csv.reader(csvFH, delimiter=',', quotechar='"')
	data = numpy.empty((258,3))
	
	i = 0
	for row in csvData:
		if row[0].find('ID') == -1:
			data[i,0] = float(row[1])
			data[i,1] = float(row[2])
			data[i,2] = float(row[3])
			i = i + 1

	csvFH.close()

	return data


def getXYZ(stands):
	"""Function to return a numpy array of the stand's x, y, and z coordinates
	(in meters) from the center pole.  If the coordinates of more than one 
	stand are needed, getXYZ can be called with a numpy array of stand 
	numbers.

	.. versionchanged:: 0.3
		Changed the numbering of the outlier from -1 to a more consistent 258
	"""

	try:
		junk = len(stands)
	except TypeError:
		stands = numpy.array([stands])

	standsXYZ = _loadPositionData()

	out = numpy.zeros((stands.shape[0],3))
	for i in range(stands.shape[0]):
		validateStand(stands[i])
		out[i,:] = standsXYZ[stands[i]-1,:]

	return out


def getRelativeXYZ(stand1, stand2):
	"""Function to return the relative coordinate difference between two 
	stands.  The output is a three-element numpy array."""

	xyzs = getXYZ(numpy.array([stand1, stand2]))
	xyz1 = numpy.squeeze(xyzs[0,:])
	xyz2 = numpy.squeeze(xyzs[1,:])

	deltaXYZ = xyz2 - xyz1

	return deltaXYZ


class PositionCache(object):
	"""PositionCache is a safe alternative to calling uvUtils.getXYZ or
	uvUtils.getRealtiveXYZ over and over again.  PositionCache loads the 
	stand positions from the included CSV file once and then uses that 
	stored information for all subsequent calls.

	.. warning::
		This assumes LWA-1 as the stations.  This will need to be changed some day.
	"""

	def __init__(self):
		"""Initialize the cache by loading in all stand positions."""

		self.standsXYZ = _loadPositionData()
		
	def __isValid(self, stand):
		if stand < 1 or stand > 258:
			raise uvUtilsError('Stand #%i is out of range (1-258)' % stand)
		
	def getXYZ(self, stands):
		"""Function to return a numpy array of the stand's x, y, and z coordinates
		(in meters) from the center pole.  If the coordinates of more than one 
		stand are needed, getXYZ can be called with a numpy array of stand 
		numbers."""

		try:
			junk = len(stands)
		except TypeError:
			stands = numpy.array([stands])
	
		out = numpy.zeros((stands.shape[0],3))
		for i in range(stands.shape[0]):
			self.__isValid(stands[i])
			out[i,:] = self.standsXYZ[stands[i]-1,:]
	
		return out
			
	def getRelativeXYZ(self, stand1, stand2):
		"""Function to return the relative coordinate difference between two 
		stands.  The output is a three-element numpy array."""

		xyzs = self.getXYZ(numpy.array([stand1, stand2]))
		xyz1 = numpy.squeeze(xyzs[0,:])
		xyz2 = numpy.squeeze(xyzs[1,:])
	
		deltaXYZ = xyz2 - xyz1
	
		return deltaXYZ
	
	def getZenithDelay(self, stand1, stand2):
		"""Return the geometrical delay in seconds for zenith as viewed by
		a baseline between stand1 and stand2."""

		import aipy
		
		b = self.getRelativeXYZ(stand1, stand2)
		top = aipy.coord.azalt2top(numpy.array([[numpy.pi/4],[numpy.pi/2]]))
		geoDelay = numpy.dot(b, top[:,0]) / c
		
		return geoDelay


def _loadDelayData(filename='lwa1-cables.csv'):
	"""Private function to load in the cable data CSV file.  Someday that file 
	should be absorbed into this script so that there is not need to keep another 
	file floating around."""

	# lwa1-cables.csv contains the stand cable lengths in meters for 256 of the
	# LWA-1 stands.  The data come from the "LWA-1 Antenna Position and Cable 
	# Data" memo, version 1.
	try:
		csvFH = open(os.path.join(dataPath, filename), 'r')	
	except IOError, (errno, strerror):
		raise uvUtilsError("%s: cable length file '%s'" % (strerror, filename))
	csvData = csv.reader(csvFH, delimiter=',', quotechar='"')
	data = numpy.empty(258)
	
	i = 0
	for row in csvData:
		if row[0].find('ID') == -1:
			data[i] = float(row[1])
			i = i + 1

	csvFH.close()

	return data


def cableDelay(stand, freq):
	"""For a given stands, return a numpy array of the cable delay in seconds 
	for a specific frequency (in Hz).  If delays for more than one frequency
	are needed, the frequencies can be passed in as a numpy array.
	"""

	# Stands start at 1, indices do not
	validateStand(stand)

	# Try this so that freq can be either a scalar or a list
	try:
		junk = len(freq)
	except TypeError:
		freq = numpy.array([freq])

	# Catch the outlier and load cable and delay information into holder variables.
	# This makes it easier to do one common set of operations with different cables.
	cableLengths = _loadDelayData()
	cableLength = cableLengths[stand-1]
	velocityFactor = 0.83
	tauD0 = 2.4e-9 # dispersion delay in seconds
	if stand == 258:
		velocityFactor = 0.85
		tauD0 = 4.6e-9	# dispersion delay in seconds

	# First, calculate the delay from cable length assuming a velocity factor.
	delay = cableLength / (velocityFactor * c)
	# Next, calculate the frequency-dependent dispersion delay and add it to 
	# the delay calculated above.  The form of the dispersion delay comes from
	# Section 4, Equation 25 of Steve Ellingson's "Dispersion in Coaxial Cables" 
	# memo.
	dispDelay = tauD0*(cableLength/100.0)/numpy.sqrt(freq/10.0e6)
	delays = delay + dispDelay

	return delays


class CableCache(object):
	"""CableCache is a safe alternative to calling uvUtils.cableDelay over 
	and over again.  CableCache loads the cable length file once and then
	uses that stored information for all subsequent calls.

	.. warning::
		This assumes LWA-1 as the stations.  This will need to be changed some day.
	"""

	def __init__(self, freq, applyDispersion=True):
		"""Initialize the cache by loading in all stand cable lengths and begin
		filling in the attributes."""

		self.cableLengths = _loadDelayData()
		try:
			junk = len(freq)
		except TypeError:
			freq = numpy.array([freq])
		self.freq = freq
		self.disp = 1.0/numpy.sqrt(numpy.abs(self.freq)/10.0e6)
		self.applyDispersion = applyDispersion
	
	def updateFreq(self, freq):
		"""Update the freq attribute of the cache to change which frequencies the
		delays are calculated for."""

		self.freq = freq
		self.disp = 1.0/numpy.sqrt(numpy.abs(self.freq)/10.0e6)

	def updateApplyDispersion(self, applyDispersion):
		"""Update the applyDispersion attribute of the cache to turn the cable
		dispersion on and off."""

		self.applyDispersion = applyDispersion

	def __isValid(self, stand):
		if stand < 1 or stand > 258:
			raise uvUtilsError('Stand #%i is out of range (1-258)' % stand)

	def cableDelay(self, stand, freq=None):
		"""For a given stands, return a numpy array of the cable delay in seconds 
		for a specific frequency (in Hz).  If delays for more than one frequency
		are needed, the frequencies can be passed in as a numpy array."""

		self.__isValid(stand)
	
		# Catch the outlier and load cable and delay information into holder variables.
		# This makes it easier to do one common set of operations with different cables.
		cableLength = self.cableLengths[stand-1]
		velocityFactor = 0.83
		tauD0 = 2.4e-9 # dispersion delay in seconds
		if stand == 258:
			velocityFactor = 0.85
			tauD0 = 4.6e-9	# dispersion delay in seconds
	
		# First, calculate the delay from cable length assuming a velocity factor.
		delay = cableLength / (velocityFactor * c)
		# Next, calculate the frequency-dependent dispersion delay and add it to 
		# the delay calculated above.  The form of the dispersion delay comes from
		# Section 4, Equation 25 of Steve Ellingson's "Dispersion in Coaxial Cables" 
		# memo.
		if freq is None:
			dispDelay = tauD0*(cableLength/100.0)*self.disp
		else:
			try:
				junk = len(freq)
			except TypeError:
				freq = numpy.array([freq])
			dispDelay = tauD0*(cableLength/100.0)/numpy.sqrt(numpy.abs(freq)/10.0e6)

		if not self.applyDispersion:
			dispDelay *= 0.0
				
		delays = delay + dispDelay

		return delays


def signalDelay(stand, freq, cache=None):
	"""Experimental function that wraps the cable delay for a stand along with 
	any other delays determined by phase fitting.  The list of stands that can be 
	corrected is incomplete.  Similar to cableDelay, a numpy array is returned.

	.. note::

		Currently all additional delays are set to 0.

	.. warning::
		This assumes LWA-1 as the stations.  This will need to be changed some day.
	"""

	# Additional delays in ns found from phase fitting.  There are currently no
	# other delays added in and signalDelay returns the same values as cableDelay.
	addDelay = {}
	#addDelay = {4: 0.0, -1: -208.6, 205: 96.5, 158: 197.2, 9: 249.8, 246: -153.4, 168: 372.4, 69: 124.4, 14: 314.9, 80: 339.7, 118: 401.3, 254: -81.1, 34: 239.7, 38: 292.5, 181: -123.6, 67: 79.2, 183: -108.6, 206: -118.2, 174: 467.5, 153: -93.2}

	# Stands start at 1, indices do not
	validateStand(stand)

	# Get the delays from the cable model and add in any additional delays found 
	# in the addDelay dictionary.
	if cache is None:
		delay = cableDelay(stand, freq)
	else:
		cache.updateFreq(freq)
		delay = cache.cableDelay(stand)
	if stand in addDelay.keys():
		delay += (addDelay[stand]*1e-9)

	return delay
	

class SignalCache(CableCache):
	"""Subclass of CableCache that does for signalDelay what CableCache did for
	cableDelay."""

	def signalDelay(self, stand):
		"""Experimental function that wraps the cable delay for a stand along with 
		any other delays determined by phase fitting.  The list of stands that can be 
		corrected is incomplete.  Similar to cableDelay, a numpy array is returned."""

		return signalDelay(stand, self.freq, cache=self)


def cableAttenuation(stand):
	"""For a given stands, return the factor needed to correct for cable 
	losses."""

	# Stands start at 1, indicies do not
	validateStand(stand)

	# Catch the outlier and load cable and delay information into holder variables.
	# This makes it easier to do one common set of operations with different cables.
	if stand == -1:
		# For the outlier, the attenuation comes from the Times Microwave LMR-400 
		# data sheet for 50 Mhz.
		cableLength = 315.344 # m
		dBperM = 0.029 # dB/m
	else:
		cableLengths = _loadDelayData()
		cableLength = cableLengths[stand-1]
		dBperM = 0.047 # dB / m

	dBLoss = dBperM*cableLength
	multFactor = 10.0**(dBLoss/10.0)

	return multFactor


def getBaselines(stands, IncludeAuto=False, Indicies=False):
	"""Generate a list of two-element tuples that describe which antennae
	compose each of the output uvw triplets from computeUVW/computeUVTrack.
	If the Indicies keyword is set to True, the two-element tuples 
	contain the indicies of the stands array used, rather than the actual
	stand numbers."""

	if IncludeAuto:
		offset = 0
	else:
		offset = 1

	try:
		N = stands.shape[0]
	except AttributeError:
		stands = numpy.array(stands)
		N = stands.shape[0]
	out = []
	for i in range(0, N-offset):
		for j in range(i+offset, N):
			if Indicies:
				out.append( (i, j) )
			else:
				out.append( (stands[i], stands[j]) )

	return out
	

def baseline2antenna(baseline, stands, BaselineList=None, IncludeAuto=False, Indicies=False):
	"""Given a baseline number, a list of stands, and options of how the base-
	line listed was generated, convert the baseline number to antenna numbers.
	Alternatively, use a list of baselines instead of generating a new list.  
	This utility is useful for figuring out what antennae comprise a baseline."""

	# Build up the list of baselines using the options provided
	if BaselineList is None:
		BaselineList = getBaselines(stands, IncludeAuto=IncludeAuto, Indicies=Indices)
	
	# Select the correct one and return based on the value of Indicies
	i,j = BaselineList[baseline]
	if Indicies:
		return i,j
	else:
		return stands[i],stands[j]


def antenna2baseline(ant1, ant2, stands, BaselineList=None, IncludeAuto=False, Indicies=False):
	"""Given two antenna numbers, a list of stands, and options to how the base-
	line listed was generated, convert the antenna pair to  a baseline number. 
	This utility is useful for picking out a particular pair from a list of
	baselines."""

	# Build up the list of baselines using the options provided
	if BaselineList is None:
		BaselineList = getBaselines(stands, IncludeAuto=IncludeAuto, Indicies=Indices)

	# If we don't want indicies, convert stand numbers to indicies
	if not Indicies:
		ant1 = (numpy.where( stands == ant1 ))[0][0]
		ant2 = (numpy.where( stands == ant2 ))[0][0]
	
	# Loop over the baselines until we find one that matches.  If we don't find 
	# one, return -1
	i = 0
	for baseline in BaselineList:
		if ant1 in baseline and ant2 in baseline:
			return i
		else:
			i = i + 1
	else:
			return -1


def computeUVW(stands, HA=0.0, dec=34.070, freq=49.0e6, IncludeAuto=False):
	"""Compute the uvw converate of a baselines formed by a collection of 
	stands.  The coverage is computed at a given HA (in hours) and 
	declination (in degrees) for LWA-1.  The frequency provided (in Hz) can 
	either as a scalar or as a numpy array.  If more than one frequency is 
	given, the output is a three dimensional with dimensions of baselines, 
	uvw, and frequencies."""

	# Try this so that freq can be either a scalar or a list
	try:
		junk = len(freq)
	except TypeError:
		freq = numpy.array([freq])

	N = stands.shape[0]
	baselines = getBaselines(stands, IncludeAuto=IncludeAuto, Indicies=True)
	Nbase = len(baselines)
	Nfreq = freq.shape[0]
	uvw = numpy.zeros((Nbase,3,Nfreq))

	# Phase center coordinates
	# Convert numbers to radians and, for HA, hours to degrees
	HA2 = HA * 15.0 * deg2rad
	dec2 = dec * deg2rad
	lat2 = 34.070 * deg2rad

	# Coordinate transformation matrices
	trans1 = numpy.matrix([[0, -math.sin(lat2), math.cos(lat2)],
					   [1, 0,               0],
					   [0, math.cos(lat2),  math.sin(lat2)]])
	trans2 = numpy.matrix([[math.sin(HA2),                 math.cos(HA2),                 0],
					   [-math.sin(dec2)*math.cos(HA2), math.sin(dec2)*math.sin(HA2),  math.cos(dec2)],
					   [math.cos(dec2)*math.cos(HA2),  -math.cos(dec2)*math.sin(HA2), math.sin(dec2)]])

	# Position cache
	posCache = PositionCache()

	count = 0
	for i,j in baselines:
		# Go from a east, north, up coordinate system to a celestial equation, 
		# east, north celestial pole system
		xyzPrime = posCache.getRelativeXYZ(stands[i], stands[j])
		xyz = trans1*numpy.matrix([[xyzPrime[0]],[xyzPrime[1]],[xyzPrime[2]]])
		
		# Go from CE, east, NCP to u, v, w
		temp = trans2*xyz
		for k in range(Nfreq):
			uvw[count,:,k] = numpy.squeeze(temp) * freq[k] / c
		count = count + 1

	return uvw


def computeUVTrack(stands, dec=34.070, freq=49.0e6):
	"""Whereas computeUVW provides the uvw coverage at a particular time, 
	computeUVTrack provides the complete uv plane track for a long 
	integration.  The output is a three dimensional numpy array with 
	dimensions baselines, uv, and 512 points along the track ellipses.  
	Unlike computeUVW, however, only a single frequency (in Hz) can be 
	specified."""

	# Try this so that freq can be either a scalar or a list
	try:
		junk = len(freq)
	except TypeError:
		freq = numpy.array([freq])

	N = stands.shape[0]
	Nbase = N*(N-1)/2
	uvTrack = numpy.zeros((Nbase,2,512))

	# Phase center coordinates
	# Convert numbers to radians and, for HA, hours to degrees
	dec2 = dec * math.pi/180.0
	lat2 = 34.070 * math.pi/180.0

	# Coordinate transformation matrices
	trans1 = numpy.matrix([[0, -math.sin(lat2), math.cos(lat2)],
					   [1, 0,               0],
					   [0, math.cos(lat2),  math.sin(lat2)]])

	# Position cache
	posCache = PositionCache()

	count = 0
	for i,j in getBaselines(stands, Indicies=True):
		# Go from a east, north, up coordinate system to a celestial equation, 
		# east, north celestial pole system
		xyzPrime = posCache.getRelativeXYZ(stands[i], stands[j])
		xyz = trans1*numpy.matrix([[xyzPrime[0]],[xyzPrime[1]],[xyzPrime[2]]])

		uRange = numpy.linspace(-math.sqrt(xyz[0]**2 + xyz[1]**2), math.sqrt(xyz[0]**2 + xyz[1]**2), num=256)
		vRange1 = numpy.sqrt(xyz[0]**2 + xyz[1]**2 - uRange**2)*sin(dec2) + xyz[2]*cos(dec2)
		vRange2 = -numpy.sqrt(xyz[0]**2 + xyz[1]**2 - uRange**2)*sin(dec2) + xyz[2]*cos(dec2)

		uvTrack[count,0,0:256] = uRange * freq / c
		uvTrack[count,1,0:256] = vRange1 * freq / c
		uvTrack[count,0,256:512] = uRange[::-1] * freq / c
		uvTrack[count,1,256:512] = vRange2[::-1] * freq / c
		count = count + 1
			
	return uvTrack



# -*- coding: utf-8 -*-

"""Unit test for lsl.misc.mathutil module."""

import unittest
import logging
import math
import StringIO

import numpy

from lsl.misc import mathutil


__revision__  = "$Revision: 88 $"
__version__   = "0.1"
__author__    = "D.L.Wood"
__maintainer__ = "Jayce Dowell"


class mathutil_tests(unittest.TestCase):
	"""A unittest.TestCase collection of unit tests for the lsl.misc.mathutil
	module."""
	
	def setUp(self):
		"""Setup for unit tests."""
		
		# disable logger since we intentionally generate errors 
		logging.basicConfig(stream = StringIO.StringIO())
	
	def test_downsample(self):
		"""Test mathutil.downsample() function."""
		
		x = numpy.array((0, 1, 2, 3))
		
		y = mathutil.downsample(x, 2, False)
		self.assertEqual(len(y), 2)
		self.assertAlmostEqual(y[0], 1.0)
		self.assertAlmostEqual(y[1], 5.0)
		
		y = mathutil.downsample(x, 2, True)
		self.assertEqual(len(y), 2)
		self.assertAlmostEqual(y[0], 0.5)
		self.assertAlmostEqual(y[1], 2.5)
		
		y = mathutil.downsample(x, 3)
		self.assertEqual(len(y), 1)
		self.assertAlmostEqual(y[0], 1.0)
		
	def test_cmagnitde(self):
		"""Test mathutil.cmagnitude() function."""
		
		x = complex(1, 0)
		self.assertAlmostEqual(mathutil.cmagnitude(x), 1.0)
		
		x = complex(0, 1)
		self.assertAlmostEqual(mathutil.cmagnitude(x), 1.0)
		
		x = complex(-1, 0)
		self.assertAlmostEqual(mathutil.cmagnitude(x), 1.0)
		
		x = complex(0, -1)
		self.assertAlmostEqual(mathutil.cmagnitude(x), 1.0)
		
		x = complex(3, 4)
		self.assertAlmostEqual(mathutil.cmagnitude(x), 5.0)
		
	def test_cphase(self):
		"""Test mathutil.cphase() function."""
		
		x = complex(1, 0)
		self.assertAlmostEqual(mathutil.cphase(x), math.radians(0.0))
		
		x = complex(0, 1)
		self.assertAlmostEqual(mathutil.cphase(x), math.radians(90.0))
		
		x = complex(-1, 0)
		self.assertAlmostEqual(mathutil.cphase(x), math.radians(180.0))
		
		x = complex(0, -1)
		self.assertAlmostEqual(mathutil.cphase(x), math.radians(-90.0))
		
	def test_cpolar(self):
		"""Test mathutil.cpolar() function."""
		
		x = numpy.array((0+1j, 1+0j, 1+1j, -1+0j))
		mag = (1.0, 1.0, math.sqrt(2.0), 1.0)
		phase = (math.radians(90.0), 0.0, math.radians(45.0), math.radians(180.0))
		im = iter(mag)
		ip = iter(phase)
		
		pol = mathutil.cpolar(x)
		for p in pol:
			self.assertAlmostEqual(p[0], im.next())
			self.assertAlmostEqual(p[1], ip.next())
			
	def test_creal(self):
		"""Test mathtutil.creal() function."""
		
		x = (1, math.radians(0.0))
		self.assertAlmostEqual(mathutil.creal(x), 1.0)
		
		x = (1, math.radians(90.0))
		self.assertAlmostEqual(mathutil.creal(x), 0.0)
		
		x = (1, math.radians(180.0))
		self.assertAlmostEqual(mathutil.creal(x), -1.0)
		
		x = (1, math.radians(-90.0))
		self.assertAlmostEqual(mathutil.creal(x), 0.0)
			
	def test_cimag(self):
		"""Test mathtutil.cimag() function."""
		
		x = (1, math.radians(0.0))
		self.assertAlmostEqual(mathutil.cimag(x), 0.0)
		
		x = (1, math.radians(90.0))
		self.assertAlmostEqual(mathutil.cimag(x), 1.0)
		
		x = (1, math.radians(180.0))
		self.assertAlmostEqual(mathutil.cimag(x), 0.0)
		
		x = (1, math.radians(-90.0))
		self.assertAlmostEqual(mathutil.cimag(x), -1.0)
		
	def test_crect(self):
		"""Test mathutil.crect() function."""
		
		x = numpy.array(((1.0, math.radians(90.0)),
					(1.0, 0.0),
					(math.sqrt(2.0), math.radians(45.0)),
					(1.0, math.radians(180.0))))
		
		c = (0+1j, 1+0j, 1+1j, -1+0j)
		ic = iter(c)
		
		rect = mathutil.crect(x)
		for r in rect:
			co = ic.next()
			self.assertAlmostEqual(r.real, co.real)
			self.assertAlmostEqual(r.imag, co.imag)
			
	def test_regrid(self):
		"""Test mathutil.regrid() function."""
		
		yout = (0.0, 2.0, 4.0, 6.0, 8.0)
		
		x = numpy.arange(0, 10, dtype = numpy.float_)
		y = numpy.arange(0, 10, dtype = numpy.float_)
		xnew = numpy.arange(0, 10, 2, dtype = numpy.float_)
		
		ynew = mathutil.regrid(x, y, xnew, method = 'spline')
		iy = iter(yout)
		for yn in ynew:
			self.assertAlmostEqual(yn, iy.next())
			
		ynew = mathutil.regrid(x, y, xnew, method = 'linear')
		iy = iter(yout)
		for yn in ynew:
			self.assertAlmostEqual(yn, iy.next())
		
		xnew = numpy.arange(-2, 10, 2, dtype = numpy.float_)    
		self.assertRaises(ValueError, mathutil.regrid, x, y, xnew)
		
		xnew = numpy.arange(0, 12, 2, dtype = numpy.float_)    
		self.assertRaises(ValueError, mathutil.regrid, x, y, xnew)
		
	def test_smooth(self):
		"""Test mathutil.smooth() function."""
		
		x = numpy.arange(0, 100, dtype = numpy.float_)
		mathutil.smooth(x)
		mathutil.smooth(x, window='flat')
		
	def test_to_dB(self):
		"""
		Test mathutil.to_dB() function.
		"""
		
		x = numpy.random.randn(100) + 1000.0
		mathutil.to_dB(x)

	def test_from_dB(self):
		"""Test mathutil.from_dB function."""
		
		x = numpy.arange(1, 100, dtype = numpy.float_)
		mathutil.from_dB(x)
		
	def test_robustmean(self):
		"""Test mathutil.robustmean() function."""
		
		x = numpy.random.randn(100)
		x[0] = 1000.0
		x[-5] = 3000.0
		
		self.assert_(mathutil.robustmean(x) < x.mean())
    
    
class mathutil_test_suite(unittest.TestSuite):
	"""A unittest.TestSuite class which contains all of the lwa_user.mathutil
	module unit tests."""
	
	def __init__(self):
		unittest.TestSuite.__init__(self)
		
		loader = unittest.TestLoader()
		self.addTests(loader.loadTestsFromTestCase(mathutil_tests))        
        
        
if __name__ == '__main__':
	unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Given a TBW file, plot the time averaged spectra for each digitizer input."""

import os
import sys
import math
import numpy
import getopt

import lsl.reader.tbw as tbw
import lsl.reader.errors as errors
import lsl.correlator.fx as fxc

import matplotlib.pyplot as plt


def usage(exitCode=None):
	print """tbwSpectra.py - Read in TBW files and create a collection of 
time-averaged spectra.

Usage: tbwSpectra.py [OPTIONS] file

Options:
-h --help                  Display this help information
-b --blackman              Apply a Blackman window to the data
-q --quiet                 Run tbwSpectra in silent mode
-l --fft-length            Set FFT length (default = 4096)
-o --output                Output file name for spectra image
"""

	if exitCode is not None:
		sys.exit(exitCode)
	else:
		return True


def parseOptions(args):
	config = {}
	# Command line flags - default values
	config['LFFT'] = 4096
	config['maxFrames'] = 300000
	config['blackman'] = False
	config['output'] = None
	config['verbose'] = True
	config['args'] = []

	# Read in and process the command line flags
	try:
		opts, args = getopt.getopt(args, "hqbl:o:", ["help", "quiet", "blackman", "fft-length=", "output="])
	except getopt.GetoptError, err:
		# Print help information and exit:
		print str(err) # will print something like "option -a not recognized"
		usage(exitCode=2)
	
	# Work through opts
	for opt, value in opts:
		if opt in ('-h', '--help'):
			usage(exitCode=0)
		elif opt in ('-q', '--quiet'):
			config['verbose'] = False
		elif opt in ('-b', '--blackman'):
			config['blackman'] = True
		elif opt in ('-l', '--fft-length'):
			config['LFFT'] = int(value)
		elif opt in ('-o', '--output'):
			config['output'] = value
		else:
			assert False
	
	# Add in arguments
	config['args'] = args

	# Return configuration
	return config

def main(args):
	# Parse command line options
	config = parseOptions(args)

	# Length of the FFT
	LFFT = config['LFFT']

	# Make sure that the file chunk size contains is an intger multiple
	# of the FFT length so that no data gets dropped
	maxFrames = int(config['maxFrames']/float(LFFT))*LFFT
	# It seems like that would be a good idea, however...  TBW data comes one
	# capture at a time so doing something like this actually truncates data 
	# from the last set of stands for the first integration.  So, we really 
	# should stick with
	maxFrames = config['maxFrames']

	fh = open(config['args'][0], "rb")
	nFrames = os.path.getsize(config['args'][0]) / tbw.FrameSize
	dataBits = tbw.getDataBits(fh)
	# The number of ant/pols in the file is hard coded because I cannot figure out 
	# a way to get this number in a systematic fashion
	antpols = 20 #2*getFramesPerObs(fh
	nChunks = int(math.ceil(1.0*nFrames/maxFrames))
	if dataBits == 12:
		nSamples = 400
	else:
		nSamples = 1200

	# File summary
	print "Filename: %s" % config['args'][0]
	print "Ant/Pols: %i" % antpols
	print "Sample Length: %i-bit" % dataBits
	print "Frames: %i" % nFrames
	print "Chunks: %i" % nChunks

	# Master loop over all of the file chuncks
	masterCount = {}
	masterWeight = numpy.zeros((nChunks, antpols, LFFT-1))
	masterSpectra = numpy.zeros((nChunks, antpols, LFFT-1))
	for i in range(nChunks):
		# Find out how many frames remain in the file.  If this number is larger
		# than the maximum of frames we can work with at a time (maxFrames),
		# only deal with that chunk
		framesRemaining = nFrames - i*maxFrames
		if framesRemaining > maxFrames:
			framesWork = maxFrames
		else:
			framesWork = framesRemaining
		print "Working on chunk %i, %i frames remaining" % (i, framesRemaining)

		count = {}
		data = numpy.zeros((antpols,framesWork*nSamples/antpols), dtype=numpy.int16)
		# If there are fewer frames than we need to fill an FFT, skip this chunk
		if data.shape[1] < LFFT:
			break
		# Inner loop that actually reads the frames into the data array
		for j in range(framesWork):
			# Read in the next frame and anticipate any problems that could occur
			try:
				cFrame = tbw.readFrame(fh, Verbose=False)
			except errors.eofError:
				break
			except errors.syncError:
				#print "WARNING: Mark 5C sync error on frame #%i" % (int(fh.tell())/TBWFrameSize-1)
				continue
			except errors.numpyError:
				break

			stand = cFrame.header.parseID()
			# In the current configuration, stands start at 1 and go up to 10.  So, we
			# can use this little trick to populate the data array
			aStand = 2*(stand-1)
			if aStand not in count.keys():
				count[aStand] = 0
				masterCount[aStand] = 0
			if cFrame.header.frameCount % 10000 == 0 and config['verbose']:
				print "%2i -> %2i  %6.3f  %5i  %i" % (stand, aStand, cFrame.getTime(), cFrame.header.frameCount, cFrame.data.timeTag)

			# Additional check on the data array bounds so that we don't overflow it.  
			# This check may be redundant...
			if (count[aStand]+1)*nSamples >= data.shape[1]:
				continue

			# Actaully load the data.  x pol goes into the even numbers, y pol into the 
			# odd numbers
			data[aStand, count[aStand]*nSamples:(count[aStand]+1)*nSamples] = numpy.squeeze(cFrame.data.xy[0,:])
			data[aStand+1, count[aStand]*nSamples:(count[aStand]+1)*nSamples] = numpy.squeeze(cFrame.data.xy[1,:])
			# Update the counters so that we can average properly later on
			count[aStand] = count[aStand] + 1
			masterCount[aStand] = masterCount[aStand] + 1

		# Calculate the spectra for this block of data and then weight the results by 
		# the total number of frames read.  This is needed to keep the averages correct.
		# NB:  The weighting is the same for the x and y polarizations because of how 
		# the data are packed in TBW
		freq, tempSpec = fxc.calcSpectra(data, LFFT=LFFT, BlackmanFilter=config['blackman'], verbose=config['verbose'])
		for stand in count.keys():
			masterSpectra[i,stand,:] = tempSpec[stand,:]
			masterSpectra[i,stand+1,:] = tempSpec[stand+1,:]
			masterWeight[i,stand,:] = count[stand]
			masterWeight[i,stand+1,:] = count[stand]

		# We don't really need the data array anymore, so delete it
		del(data)

	# Now that we have read through all of the chunks, peform the final averaging by
	# dividing by all of the chunks
	spec = numpy.squeeze( (masterWeight*masterSpectra).sum(axis=0) / masterWeight.sum(axis=0) )

	# The plots:  This is setup for the current configuration of 20 antpols
	fig = plt.figure()
	figsX = int(round(math.sqrt(antpols)))
	figsY = antpols / figsX
	for i in range(antpols):
		ax = fig.add_subplot(figsX,figsY,i+1)
		currSpectra = numpy.squeeze( numpy.log10(spec[i,:])*10.0 )
		ax.plot(freq/1e6, currSpectra, label='%i' % (i+1))

		# If there is more than one chunk, plot the difference between the global 
		# average and each chunk
		if nChunks > 1:
			for j in range(nChunks):
				# Some files are padded by zeros at the end and, thus, carry no 
				# weight in the average spectra.  Skip over those.
				if masterWeight[j,i,:].sum() == 0:
					continue

				# Calculate the difference between the spectra and plot
				subspectra = numpy.squeeze( numpy.log10(masterSpectra[j,i,:])*10.0 )
				diff = subspectra - currSpectra
				ax.plot(freq/1e6, diff)

		ax.set_title('Input: %i' % (i+1))
		ax.set_xlabel('Frequency [MHz]')
		ax.set_ylabel('P. S. [Arb. dB]')
		ax.set_xlim([10,90])

	plt.show()

	# Save spectra image if requested
	if config['output'] is not None:
		fig.savefig(config['output'])


if __name__ == "__main__":
	main(sys.argv[1:])



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""Application to calculate real-time ephemeris for a LWA site."""

import curses
import math
import time
import traceback
import sys, os, string
import optparse

from lsl import astro
from lsl import transform
from lsl.common import stations as lwa_common

__revision__  = "$Revision: 94 $"
__version__   = "dev"
__author__    = "P.S.Ray"


def restorescreen():
    curses.nocbreak()
    curses.echo()
    curses.endwin()
    

display = string.Template("""==========================================================================
UTC : $utc          MJD(UTC) $mjdutc
Site: $loc                              LST: $lst
Long: $long                    Lat: $lat
==========================================================================
----------- Sun --------------          ------------ Jupiter -------------
RA(Cur)          $sun_ra        RA(Cur)               $jup_ra
Dec(Cur)        $sun_dec         Dec(Cur)             $jup_dec
Alt                   $sun_alt deg        Alt                  $jup_alt deg
Az                    $sun_az deg        Az                   $jup_az deg
---- Galactic Center ---------          ---------- Tau A (Crab) ----------
RA(Cur)          $gal_ra        RA(Cur)               $tau_ra
Dec(Cur)        $gal_dec         Dec(Cur)             $tau_dec
Alt                   $gal_alt deg        Alt                  $tau_alt deg
Az                    $gal_az deg        Az                   $tau_az deg
---------- Cas A -------------          ------------ Cyg A ---------------
RA(Cur)          $cas_ra        RA(Cur)               $cyg_ra
Dec(Cur)        $cas_dec         Dec(Cur)             $cyg_dec
Alt                   $cas_alt deg        Alt                   $cyg_alt deg
Az                    $cas_az deg        Az                    $cyg_az deg
==========================================================================""")


if __name__ == '__main__':
	usage = "astrostatus [options]"

	# check command line
	parser = optparse.OptionParser(usage = usage)
	parser.add_option("-s", "--site", action = "store", type = "string",
					dest = "site", default = 'LWA-1', help = "site name (default LWA-1)")
	(opts, args) = parser.parse_args()
			
	# setup transform objects
	lwa1 = lwa_common.lwa1()
	site = transform.GeographicalPosition((lwa1.long*180.0/math.pi, lwa1.lat*180.0/math.pi), name=lwa1.name)
	
	sun_pos = transform.PlanetaryPosition('Sun')
	sun_dir = transform.PointingDirection(sun_pos, site)
	
	jup_pos = transform.PlanetaryPosition('Jupiter')
	jup_dir = transform.PointingDirection(jup_pos, site)
	
	gal_pos = transform.CelestialPosition((astro.hms(17, 42, 0.1),
	astro.dms(True, 28, 55, 8)), name = 'SgrA')
	gal_dir = transform.PointingDirection(gal_pos, site)
	
	tau_pos = transform.CelestialPosition((astro.hms(5, 34, 0.5),
	astro.dms(False, 22, 0, 52)), name = 'TauA')
	tau_dir = transform.PointingDirection(tau_pos, site)
	
	cyg_pos = transform.CelestialPosition((astro.hms(19, 59, 0.8),
	astro.dms(False, 40, 44, 2)), name = 'CygA')
	cyg_dir = transform.PointingDirection(cyg_pos, site)
	
	cas_pos = transform.CelestialPosition((astro.hms(23, 23, 0.7),
	astro.dms(False, 58, 49, 16)), name = 'CasA')
	cas_dir = transform.PointingDirection(cas_pos, site)
    
	try:
		# setup display screen
		screen = curses.initscr()
		curses.noecho()
		curses.cbreak()
		screen.nodelay(1)

		strdict = {}
		while (True):
			screen.clear()

			# Build string to print
			# update current time and local sidereal time
			currentTime = transform.Time.from_system() 
						
			strdict['utc'] = "%-25s" % currentTime.utc_str
			strdict['mjdutc'] = "%.5f" % currentTime.utc_mjd
			strdict['loc'] = site.name
			(lng, lat) = site.geo.format()
			strdict['lat'] = lat
			strdict['long'] = "%-15s" % lng
			strdict['lst'] = astro.deg_to_dms(site.sidereal(currentTime))
			
			# calculate Sun position 
			sun_hrz = sun_dir.hrz(currentTime)
			(ra, dec) = sun_pos.apparent_equ(currentTime).format()
			strdict['sun_ra'] = "%-15s" % ra
			strdict['sun_dec'] = "%-15s" % dec
			strdict['sun_alt'] = "%6.2f" % sun_hrz.alt
			strdict['sun_az'] = "%6.2f" % sun_hrz.az

			# calculate Jupiter posn
			jup_hrz = jup_dir.hrz(currentTime)
			(ra, dec) = jup_pos.apparent_equ(currentTime).format()
			strdict['jup_ra'] = ra
			strdict['jup_dec'] = dec
			strdict['jup_alt'] = "%6.2f" % jup_hrz.alt
			strdict['jup_az'] = "%6.2f" % jup_hrz.az

			# Galactic Center
			gal_hrz = gal_dir.hrz(currentTime)
			(ra, dec) = gal_pos.apparent_equ(currentTime).format()
			strdict['gal_ra'] = "%-15s" % ra
			strdict['gal_dec'] = "%-15s" % dec
			strdict['gal_alt'] = "%6.2f" % gal_hrz.alt
			strdict['gal_az'] = "%6.2f" % gal_hrz.az

			# calculate Tau A (Crab) position
			tau_hrz = tau_dir.hrz(currentTime)
			(ra, dec) = tau_pos.apparent_equ(currentTime).format()
			strdict['tau_ra'] = ra
			strdict['tau_dec'] = dec
			strdict['tau_alt'] = "%6.2f" % tau_hrz.alt
			strdict['tau_az'] = "%6.2f" % tau_hrz.az

			# calculate Cygnus A position
			cyg_hrz = cyg_dir.hrz(currentTime)
			(ra, dec) = cyg_pos.apparent_equ(currentTime).format()
			strdict['cyg_ra'] = ra
			strdict['cyg_dec'] = dec
			strdict['cyg_alt'] = "%6.2f" % cyg_hrz.alt
			strdict['cyg_az'] = "%6.2f" % cyg_hrz.az

			# calculate Cassiopeia A position
			cas_hrz = cas_dir.hrz(currentTime)
			(ra, dec) = cas_pos.apparent_equ(currentTime).format()
			strdict['cas_ra'] = "%-15s" % ra
			strdict['cas_dec'] = "%-15s" % dec
			strdict['cas_alt'] = "%6.2f" % cas_hrz.alt
			strdict['cas_az'] = "%6.2f" % cas_hrz.az
		
			# Refresh screen and wait 1 second
			screen.addstr(0,0,display.safe_substitute(strdict))
			screen.refresh()
			time.sleep(1)
				
			# Check for keypress and exit if Q
			c = screen.getch()
			if (c > 0):
				if chr(c) == 'q': 
					break
				if chr(c) == 'Q': 
					break
			
		restorescreen()
		
	except KeyboardInterrupt:
		restorescreen()



# -*- coding: utf-8 -*-

"""Unit test for lsl.misc.geodesy module."""

import os
import unittest

from lsl.misc import geodesy


__revision__ = "$Revision:1 $"
__version__  = "0.1"
__author__    = "Jayce Dowell"


class geodesy_tests(unittest.TestCase):
	"""A unittest.TestCase collection of unit tests for the lsl.common.paths tests
	module."""

	def test_read_MAIA(self):
		"""Test that a line from one of the MAIA files can be read."""

		ACCURACY = 6

		line = '92 1 1 48622.00 I  0.182969 0.000672  0.168817 0.000345  I-0.1251669 0.0000207  1.8335 0.0201  I   -10.437     .507     -.917     .165   .182400   .167900  -.1253000    -9.900    -1.700\n'
		newEOP = geodesy.EOP()
		newEOP.fromMAIA(line)

		self.assertAlmostEqual(newEOP.x, 0.182969, ACCURACY)
		self.assertAlmostEqual(newEOP.y, 0.168817, ACCURACY)
		self.assertAlmostEqual(newEOP.utDiff, -0.1251669, ACCURACY)
		self.assertEqual(newEOP.type, 'final')

	def test_read_mjd(self):
		"""Test reading a specific MJD via getEOP."""

		ACCURACY = 6

		# Read in the data for January 2, 1973
		eops = geodesy.getEOP(41684.0)
		eop = eops[0]

		# Check
		self.assertAlmostEqual(eop.x, 0.120728, ACCURACY)
		self.assertAlmostEqual(eop.y, 0.137027, ACCURACY)
		self.assertAlmostEqual(eop.utDiff,  0.8084209, ACCURACY)

		# Read in three values and check the output
		mjdList = [41684.0, 41685.0, 41686.0]
		eops = geodesy.getEOP(mjdList)
		self.assertEqual(len(eops), 3)
		for i in range(3):
			self.assertTrue(mjdList[i] in eops)

	def test_read_range(self):
		"""Test that the rangeing in getEOPRange is correct."""

		eops = geodesy.getEOPRange(start=55500.0, stop=55505.0)
		self.assertEqual(len(eops), 6)
		for i in range(55500, 55505+1):
			self.assertTrue(i in eops)

	# This next test is extermely slow on some systems, skip it.
	#def test_read_historic(self):
		#"""Test reading in the historic database."""

		#ACCURACY = 6

		## Read in the time range covered by the file
		#eops = geodesy.getEOPRange(start=41684.0, stop=55518.0)
		
		## Check the first date
		#self.assertTrue(41684 in eops)
		#eop = eops[eops.index(41684)]
		#self.assertAlmostEqual(eop.x, 0.120728, ACCURACY)
		#self.assertAlmostEqual(eop.y, 0.137027, ACCURACY)
		#self.assertAlmostEqual(eop.utDiff,  0.8084209, ACCURACY)

		## Check the last date
		#self.assertTrue(55518 in eops)
		#eop = eops[eops.index(55518)]
		#self.assertAlmostEqual(eop.x, 0.207125, ACCURACY)
		#self.assertAlmostEqual(eop.y, 0.263491, ACCURACY)
		#self.assertAlmostEqual(eop.utDiff, -0.1099045, ACCURACY)

		## Check for the presence of other dates
		#self.assertFalse(48600 in eops)


class geodesy_test_suite(unittest.TestSuite):
	"""A unittest.TestSuite class which contains all of the lsl.misc.geodesy
	module unit tests."""
	
	def __init__(self):
		unittest.TestSuite.__init__(self)
		
		loader = unittest.TestLoader()
		self.addTests(loader.loadTestsFromTestCase(geodesy_tests)) 


if __name__ == '__main__':
	unittest.main()



# -*- coding: utf-8 -*-

"""Python module for reading data in from S60 files.  Unlike TBW or TBN data, 
the S60 data consists of only packed binary values.  Thus, the functions for 
reading in the S60 data do not return objects describing each frame but rather
return numpy arrays of the data.  The two reading functions defined are:

readFrame
  return a complex array containing the data from one UDP packet (734 samples)

readChunk
  return a complex array containing a specified number of samples, i.e., the 
  'chunk' size.

In addition to the two readers, there is also a function, getBandpassModel, 
which reads in Joe Craig's model of the S60 bandpass. See `the S60 data page <http://www.phys.unm.edu/~lwa/dokuwiki/doku.php?id=s60_data#view_and_plot_with_octave>`_
for information about the model."""

import numpy

from lsl.common.paths import data as dataPath
from errors import numpyError

__version__ = '0.1'
__revision__ = '$ Revision: 2 $'
__all__ = ['readFrame', 'readChunk', 'getBandpassModel', 'FrameSize', 'SampleRate', '__version__', '__revision__', '__all__']

# Packet length seems to describe how many bytes of data are in 
# each UDP packet.  1468 bytes == 734 samples.
FrameSize = 1468
# Packets per second seens to indicate how many UDP packets are 
# sent out by the S60 system to be recorded every second.  Thus in
# one second, it would seem that there are 5109*1468 = 7.50 M byes 
# or 3.75 M samples.  
__PacketsPerSecond = 5109
# Sample rate would then be the product of FrameSize and 
# PacketsPerSecond / 2 (two bytes used per I/Q sample)
SampleRate = FrameSize*__PacketsPerSecond / 2.0


def readFrame(filehandle):
	"""Function to read in a single S60 frame and return the data as a numpy
	single-precision complex array with 734 elements."""

	rawData = numpy.fromfile(filehandle, dtype=numpy.uint8, count=FrameSize)
	data = numpy.zeros(FrameSize/2, dtype=numpy.csingle)
	if rawData.shape[0] < 2*data.shape[0]:
		raise numpyError()

	data.real = rawData[0::2] - 128.0
	data.imag = rawData[1::2] - 128.0

	return data


def readChunk(filehandle, Chunk=4096):
	"""Function to read a certain number of chunks (samples) from the input file-
	handle and return them as a single-precision complex numpy array.  The output 
	is in samples so twice as many bytes need to be read in as samples."""

	rawData = numpy.fromfile(filehandle, dtype=numpy.uint8, count=2*Chunk)
	data = numpy.empty(Chunk, dtype=numpy.csingle)
	if rawData.shape[0] < 2*data.shape[0]:
		raise numpyError()
	
	data.real = rawData[0::2] - 128.0
	data.imag = rawData[1::2] - 128.0

	return data


def getBandpassModel():
	"""Read in Joe's model for the S60 badpass that is posted on the wiki."""

	from scipy.io import loadmat

	modelFile = os.path.join(dataPath, 's60-bandpass-model.mat')
	model = loadmat(modelFile, struct_as_record=True, squeeze_me=True)
	return model['y']



# -*- coding: utf-8 -*

"""lsl.common - Common information for the LSL package.  Including:
 * warns    - depercation and experimental feature warnings, 
 * paths    - paths to the module and its data directories, 
 * dp       - information about the DP system, and 
 * stations - information about LWA stations."""

import warns

import paths
import dp
import stations



# -*- coding: utf-8 -*-

"""
LWA astronomical source catalogs.
"""

# $Id: catalog.py 84 2010-05-19 22:24:00Z dwood $



import os
import math
import abc
import collections
import logging

from lsl import astro
from lsl import transform
from lsl.common.paths import data as dataPath



__revision__  = "$Revision: 84 $"
__version__   = "dev"
__author__    = "D.L.Wood"
__maintainer__ = "Jayce Dowell"



class CatalogEntry(object):
  """
  Represents one source entry in a catalogue.
  
  Contains members:
    name        - The source name.
    position    - The source equatorial J2000 position as object
                  of type transform.CelestialPosition.
    alias_list  - A list of strings providing alternate names for
                  the source.
  """
  
  
  # limit the class attributes
  
  __slots__ = ('name', 'position', 'alias_list')
    
    
  def __init__(self, name, position):
    """
    Create a catalog entry.
    """
    
    self.name = name
    self.position = position
    self.alias_list = []


  def __repr__(self):
    """
    Low level string representation.
    """
    
    return "%s.%s(%s,%s)" % (type(self).__module__, type(self).__name__,
      repr(self.name), repr(self.position))




class Catalog(collections.Mapping):
  """
  Class representing astronomical source catalog information.
  This is an abstract class; derived classes must provide a
  parse_file() method which populates the catalog object with
  information from file or other source.
  
  Catalog instances support the read-only collections.Mapping
  interface.  That is, they support the read-only methods of
  the dict built-in type.
  """
  
  __metaclass__ = abc.ABCMeta
  
  # a logging instance for catalogs
  
  log = logging.getLogger('catalog')
        

  def __init__(self, name):
    """
    Create a source catalog.
    """

    
    # initialize catalog data structures
      
    self.name = name
    self.source_map = {}
    self.alias_map = {}
       
    # parse_file() is an abstract method which must be defined in
    # a concrete implementation for a particular catalog
    
    self.log.debug("parsing data file for catalog %s", name)    
    self.parse_file()
    
    
  @abc.abstractmethod
  def parse_file(self):
    """
    Read catalog information from file into internal data
    structures.
    """
    
    pass
    
    
  @staticmethod
  def get_directory():
    """
    Returns the path to the catalog data file directory.
    """
        
    return os.path.join(dataPath, 'catalog')
        
    
  def __repr__(self):   
    """
    Low level string representation.
    """
      
    return "%s.Catalog(%s)" % (type(self).__module__, repr(self.name))
    
    
  def __len__(self):
    """
    Return the number of sources in the catalog.
    """
    
    return len(self.source_map)
    
    
  def __getitem__(self, key):
    """
    Access source by subscript name.  Raises KeyError if the source
    is not in the catalog.
    """
    
    entry = self.lookup(key)
    if entry is None:
      raise KeyError("name %s not in catalog" % repr(key))
    return entry
    
    
  def __iter__(self):
    """
    Return an iterator to the primary names in the catalog.
    """
    
    return iter(self.source_map.keys())


  def lookup(self, name):
    """
    Lookup a source in the catalog.
    
    Param: name - The primary name or alias of the source.
    
    Returns: An object of type CatalogEntry giving the source information,
             or None if the name is not found in the catalog.
    """

    try:
      entry = self.source_map[name]
    except KeyError:
      try:
        entry = self.alias_map[name]
      except KeyError:
        entry = None
            
    return entry
    
        
        

class LWA_Catalog(Catalog):
  """
  Specific definition for LWA observation source catalogue data file.
  """         

    
  def __init__(self):
    """
    Create a LWA catalog instance.
    """
    
    Catalog.__init__(self, 'LWA')
    
    
  def parse_file(self):
    """
    Read a source catalog data file.
    """

    # open data file 

    fileName = os.path.join(self.get_directory(), 'lwa_catalog.dat')
    catFile = open(fileName, 'r')
        
    # read source info
    
    lineNum = 0

    for line in catFile:
    
      lineNum += 1
      if line.startswith('#') or line.isspace():
        continue
                
      try:
      
        name = line[0:8]
        raHours = int(line[9:11], 10)
        raMinutes = int(line[12:14], 10)
        raSeconds = float(line[15:20])
        decSign = line[21]
        decDegrees = int(line[22:24], 10)
        decMinutes = int(line[25:27], 10)
        decSeconds = float(line[28:32])
      
      except ValueError as err:
        
        raise RuntimeError("file %s, line %d incorectly formated: \'%s\' : %s]" % \
          (fileName, lineNum, line, err))
                    
      name = name.rstrip()           

      ra = astro.hms(raHours, raMinutes, raSeconds)
            
      if decSign == '-':
        sign = True
      else:
        sign = False
      dec = astro.dms(sign, decDegrees, decMinutes, decSeconds)
            
      entry = CatalogEntry(name, 
        transform.CelestialPosition((ra, dec), name = name))
      self.source_map[name] = entry

      aliasList = line[34:].split()
      if len(aliasList) > 0:
        for alias in aliasList:
          entry.alias_list.append(alias)
          self.alias_map[alias] = entry
         
    catFile.close()




class PSR_Catalog(Catalog):
    """
    Specific definition for ATNF Pulsar (PSRCAT) catalog.
    Data file is psrcat.db which can be retreived from:
    <http://www.atnf.csiro.au/research/pulsar/psrcat/download.html>
    """ 
    
    
    def __init__(self):
        """
        Create an instance of the PSR catalog.
        """
        
        Catalog.__init__(self, 'PSR')
        

    def parse_file(self):
        """
        Read a source catalog data file.
        """

        debug = False
        # open data file 

        fileName = os.path.join(self.get_directory(), 'psrcat.db')
        catFile = open(fileName, 'r')
        
        # read source info
        psrb = None
        psrj = None
        ra = None
        dec = None
        bad = False

        for line in catFile:
            if line.startswith('PSRB'):
                psrb = line.split()[1]
            if line.startswith('PSRJ'):
                psrj = line.split()[1]
            if line.startswith('RAJ'):
                rastr = line.split()[1]
                sp = rastr.split(':')
                if len(sp) == 3:
                    (raHours, raMinutes, raSeconds) = sp
                elif len(sp) == 2:
                    (raHours, raMinutes) = sp
                    raSeconds = 0.0
                else:
                    if debug: print('Bad format for RAJ line : '+line)
                    bad = True
                raHours = int(raHours)
                raMinutes = int(raMinutes)
                raSeconds = float(raSeconds)
                try:
                    ra = astro.hms(raHours, raMinutes, raSeconds)
                except:
                    if debug: print('PSRCAT: Bad RA for ',psrj," : ",rastr)
                    bad = True
            if line.startswith('DECJ'):
                decstr = line.split()[1]
                sp = decstr.split(':')
                if len(sp) == 3:
                    (decDegrees, decMinutes, decSeconds) = sp
                elif len(sp) == 2:
                    (decDegrees, decMinutes) = sp
                    decSeconds = 0.0
                else:
                    if debug: print('PSRCAT: Bad format for DECJ line : '+line)
                    bad = True
                    continue
                if decDegrees.startswith('-'):
                    decDegrees = decDegrees[1:]
                    sign = True
                else:
                    sign = False
                decDegrees = int(decDegrees)
                decMinutes = int(decMinutes)
                decSeconds = float(decSeconds)
                try:
                    dec = astro.dms(sign, decDegrees, decMinutes, decSeconds)
                except:
                    if debug: print('PSRCAT: Bad DEC for ',psrj," : ",decstr)
                    bad = True
                
            if line.startswith('@-'):
                # New source, save current source info
                if psrb is not None:
                    name = psrb
                    alias = psrj
                else:
                    name = psrj
                    alias = None

                if ra is None or dec is None:
                    # These pulsars don't have RAJ, DECJ
                    # I think they may all have ecliptic positions
                    # which should be converted to ra,dec but I'm
                    # going to ignore them for now. -- paulr
                    #print "PSRCAT: No position for pulsar ",name
                    bad = True

                # Add source to list if good.
                if not bad:                    
                    sourcePos = astro.equ_posn(ra, dec) 
                    entry = CatalogEntry(name, 
                      transform.CelestialPosition(sourcePos, name = name))
                    self.source_map[name] = entry
                    if debug: print('Added ',name)
    
                    if alias is not None:
                        alias = alias.rstrip()
                        self.alias_map[alias] = entry
                        entry.alias_list.append(alias)
                        if debug: print('Alias : ',alias.rstrip())
                    
                # Clear out vars for next source
                psrb = None
                psrj = None
                ra = None
                dec = None
                bad = False

        catFile.close()

    


class PKS_Catalog(Catalog):
    """
    Specific definition for PKS source catalog.
    """


    def __init__(self):
        """
        Create a PKS catalog instance.
        """
        
        Catalog.__init__(self, 'PKS')
        
        

    def parse_file(self):
        """
        Read a source catalog data file.
        """

        # open data file 

        fileName = os.path.join(self.get_directory(), 'pkscat.txt')
        catFile = open(fileName, 'r')
        
        # read source info

        lineNum = 1
        line = catFile.readline()
        while line:

            if line == '\n':
                line = catFile.readline()
                lineNum += 1
                continue

            try:
                name = line[0:8]
                alias = line[12:19]
                raHours = int(line[22:24])
                raMinutes = int(line[25:27])
                raSeconds = float(line[28:32])
                decSign = line[33]
                decDegrees = int(line[34:36])
                decMinutes = int(line[37:39])
                decSeconds = int(line[40:42])
            except ValueError:
                raise RuntimeError("file %s, line %d incorectly formated [%s]" % \
                    (fileName, lineNum, line))   

            ra = astro.hms(raHours, raMinutes, raSeconds)
            
            if decSign == '-':
                sign = True
            else:
                sign = False
            dec = astro.dms(sign, decDegrees, decMinutes, decSeconds)
            
            sourcePos = astro.equ_posn(ra, dec)
            
            # precess coordinates from B1950
            
            entry = CatalogEntry(name, 
              transform.CelestialPosition(sourcePos, epoch = 'B1950', name = name))
            self.source_map[name] = entry

            if len(alias.strip()):
                alias = alias.rstrip()
                self.alias_map[alias] = entry
                entry.alias_list.append(alias)

            line = catFile.readline()
            lineNum += 1
            
        catFile.close()
        



class PKS90_Catalog(Catalog):
    """
    Specific definition for PKS90 source catalogue data file.
    """

    
    def __init__(self):
        """
        Create a PKS90 catalog instance.
        """
        
        Catalog.__init__(self, 'PKS90')
  

    def parse_file(self):
        """
        Read a source catalog data file.
        """

        # open data file 

        fileName = os.path.join(self.get_directory(), 'PKS90.txt')
        catFile = open(fileName, 'r')
        
        # read source info

        lineNum = 1
        line = catFile.readline()
        while line:

            if line == '\n':
                line = catFile.readline()
                lineNum += 1
                continue

            try:
                name = line[0:10]
                alias0 = line[139:148]
                alias1 = line[168:175]
                raHours = int(line[10:12])
                raMinutes = int(line[13:15])
                raSeconds = float(line[16:20])
                decSign = line[23]
                decDegrees = int(line[24:26])
                decMinutes = int(line[27:29])
                decSeconds = float(line[30:34])
            except ValueError:
                raise RuntimeError("file %s, line %d incorectly formated [%s]" % \
                    (fileName, lineNum, line))
                    
            ra = astro.hms(raHours, raMinutes, raSeconds)
            
            if decSign == '-':
                sign = True
            else:
                sign = False
            dec = astro.dms(sign, decDegrees, decMinutes, decSeconds)
            
            sourcePos = astro.equ_posn(ra, dec)
            entry = CatalogEntry(name, 
              transform.CelestialPosition(sourcePos, name = name))
            self.source_map[name] = entry

            if len(alias0.strip()): 
                alias = alias0.rstrip()
                entry.alias_list.append(alias)
                self.alias_map[alias] = entry
            if len(alias1.strip()):
                alias = alias1.rstrip()
                entry.alias_list.append(alias)
                self.alias_map[alias] = entry

            line = catFile.readline()
            lineNum += 1
            
        catFile.close()




class C3C_Catalog(Catalog):
    """
    Specific definition for Cambridge 3C source catalogue data file.
    """ 

    
    def __init__(self):
        """
        Create a 3C catalog instance.
        """
        
        Catalog.__init__(self, '3C')
        

    def parse_file(self):
        """
        Read a source catalog data file.
        """

        # open data file 

        fileName = os.path.join(self.get_directory(), '3c.dat')
        catFile = open(fileName, 'r')
        
        # read source info

        lineNum = 1
        line = catFile.readline()
        while line:

            try:
                name = line[0:3]
                raHours = int(line[12:14])
                raMinutes = int(line[15:17])
                raSeconds = float(line[18:22])
                decSign = line[27]
                decDegrees = int(line[28:30])
                decMinutes = float(line[31:35])
            except ValueError:
                raise RuntimeError("file %s, line %d incorectly formated [%s]" % \
                    (fileName, lineNum, line))
                    
            name = ('3C' + name.strip())         
                    
            ra = astro.hms(raHours, raMinutes, raSeconds)
            
            (decSeconds, decMinutes) = math.modf(decMinutes)
            if decSign == '-':
                sign = True
            else:
                sign = False
            dec = astro.dms(sign, decDegrees, int(decMinutes), (60 * decSeconds))
            
            sourcePos = astro.equ_posn(ra, dec)
            
            # precess coordinates from B1950
            
            entry = CatalogEntry(name, 
              transform.CelestialPosition(sourcePos, epoch = 'B1950', name = name))
            self.source_map[name] = entry        
                    
            line = catFile.readline()
            lineNum += 1      
                      
        catFile.close()



class C4C_Catalog(Catalog):
    """
    Specific definition for Cambridge 4C source catalogue data file.
    """ 

    def __init__(self):
        """
        Create a 4C catalog instance.
        """
        
        Catalog.__init__(self, '4C')
        

    def parse_file(self):
        """
        Read a source catalog data file.
        """
        
        # open data file 

        fileName = os.path.join(self.get_directory(), '4c.dat')
        catFile = open(fileName, 'r')
        
        # read source info

        lineNum = 1
        line = catFile.readline()
        while line:

            try:
                name = line[0:8]
                raHours = int(line[10:12])
                raMinutes = int(line[13:15])
                raSeconds = float(line[16:20])
                decSign = line[22]
                decDegrees = int(line[23:25])
                decMinutes = float(line[26:30])
                alias = line[64:-1]
            except ValueError:
                raise RuntimeError("file %s, line %d incorectly formated [%s]" % \
                    (fileName, lineNum, line))
            
            name = name.strip()        
            name = ('4C' + name)
            
            alias = alias.strip()
            alias = alias.rstrip('*')         
                    
            ra = astro.hms(raHours, raMinutes, raSeconds)
            
            (decSeconds, decMinutes) = math.modf(decMinutes)
            if decSign == '-':
                sign = True
            else:
                sign = False
                
            dec = astro.dms(sign, decDegrees, int(decMinutes), (60 * decSeconds))
            
            sourcePos = astro.equ_posn(ra, dec)
            
            # precess coordinates from B1950
           
            entry = CatalogEntry(name, 
              transform.CelestialPosition(sourcePos, epoch = 'B1950', name = name))
            self.source_map[name] = entry
            
            if len(alias.strip()):
                alias = alias.rstrip()
                entry.alias_list.append(alias)
                self.alias_map[alias] = entry       
                    
            line = catFile.readline()
            lineNum += 1      
                      
        catFile.close()
        
        

class F1FGL_Catalog(Catalog):
  """
  Specific definition for Fermi LAT first point source catalog.
  """
  
  
  def __init__(self):
    """
    Create a 1FGL catalog instance.
    """
    
    Catalog.__init__(self, '1FGL')
    
    
  def parse_file(self):
    """
    Read a source catalogue data file.
    """
    
    import pyfits
    
    # open data file
    
    fileName = os.path.join(self.get_directory(), 'gll_psc_v02.fit')
    catFile = pyfits.open(fileName)
    
    # read source info
    
    sourceTable = catFile['LAT_POINT_SOURCE_CATALOG'].data
    
    for row in sourceTable:
      
      name = str(row.field('Source_Name')).replace(' ', '_')
      ra = float(row.field('RA'))
      dec = float(row.field('DEC'))
      entry = CatalogEntry(name, 
        transform.CelestialPosition((ra, dec), name = name))
      self.source_map[name] = entry
      
      alias = str(row.field('0FGL_Name'))
      if len(alias):
        alias = alias.replace(' ', '_')
        self.alias_map[alias] = entry
        entry.alias_list.append(alias)
      
      alias = str(row.field('ASSOC_GAM1'))
      if len(alias):
        alias = alias.replace(' ', '_')
        self.alias_map[alias] = entry
        entry.alias_list.append(alias)
      
      alias = str(row.field('ASSOC_GAM2'))
      if len(alias):
        alias = alias.replace(' ', '_')
        self.alias_map[alias] = entry
        entry.alias_list.append(alias)
      
      alias = str(row.field('ASSOC1'))
      if len(alias):
        alias = alias.replace(' ', '_')
        self.alias_map[alias] = entry
        entry.alias_list.append(alias)
        
      alias = str(row.field('ASSOC2'))
      if len(alias):
        alias = alias.replace(' ', '_')
        self.alias_map[alias] = entry
        entry.alias_list.append(alias)
                
    catFile.close()



class CatalogFactory(object):
  """
  Get catalog objects by name.  Caches the catalog data so that
  the data file is parsed only once per session.
  """
  
  # a mapping of catalog names to classes
  
  catalog_class_map = \
  {
    'LWA'   : LWA_Catalog,
    'PSR'   : PSR_Catalog,
    'PKS'   : PKS_Catalog,
    'PKS90' : PKS90_Catalog,
    '3C'    : C3C_Catalog,
    '4C'    : C4C_Catalog,
    '1FGL'  : F1FGL_Catalog,
  }
  
  # a mapping of catalog names to instances
  
  catalog_instance_map = {}
  
  
  @classmethod
  def get_catalog(klass, name):
    """
    Returns a Catalog object representing the catalog
    given by name.
    """
    
    # check parameters
    
    if name not in list(klass.catalog_class_map.keys()):
      raise ValueError("unknown catalog \'%s\'" % name)
    
    # try to find an already created instance
    # if not found, create a new instance and cache
    
    try:
      catalog = klass.catalog_instance_map[name]
    except KeyError:
      catalogClass = klass.catalog_class_map[name]
      catalog = catalogClass()
      klass.catalog_instance_map[name] = catalog
      
    return catalog
    

  @classmethod
  def get_names(klass):
    """
    Return a list of known catalog names.
    """
    
    return list(klass.catalog_class_map.keys())
    




# -*- coding: utf-8 -*-

"""Small collection of robust statistical estimators based on functions from
the AstroIDL User's Library.  Function included are:
  * robustMean - robust estimator of the mean of a data set and
  * robustSigma - robust estimator of the standard deviation of a data set.
"""

import math
import numpy

__version__ = '0.1'
__revision__ = '$ Revision: 2 $'
__all__ = ['robustMean', 'robustSigma', '__version__', '__revision__', '__all__']

__epsilon = 1.0e-20

def robustMean(inputData, Cut=3.0):
	"""Robust estimator of the mean of a data set.  Based on the 
	resistant_mean function from the AstroIDL User's Library.

	.. seealso::
		:func:`lsl.misc.mathutil.robustmean`
	"""

	data = inputData.ravel()

	data0 = numpy.median(data)
	maxAbsDev = numpy.median(numpy.abs(data-data0)) / 0.6745
	if maxAbsDev < __epsilon:
		maxAbsDev = (numpy.abs(data-data0)).mean() / 0.8000

	cutOff = Cut*maxAbsDev
	good = numpy.where( numpy.abs(data-data0) <= cutOff )
	good = good[0]
	dataMean = data[good].mean()
	dataSigma = math.sqrt( ((data[good]-dataMean)**2.0).sum() / len(good) )

	if Cut > 1.0:
		sigmaCut = Cut
	else:
		sigmaCut = 1.0
	if sigmaCut <= 4.5:
		dataSigma = dataSigma / (-0.15405 + 0.90723*sigmaCut - 0.23584*sigmaCut**2.0 + 0.020142*sigmaCut**3.0)

	cutOff = Cut*dataSigma
	good = numpy.where(  numpy.abs(data-data0) <= cutOff )
	good = good[0]
	dataMean = data[good].mean()
	if len(good) > 3:
		dataSigma = math.sqrt( ((data[good]-dataMean)**2.0).sum() / len(good) )

	if Cut > 1.0:
		sigmaCut = Cut
	else:
		sigmaCut = 1.0
	if sigmaCut <= 4.5:
		dataSigma = dataSigma / (-0.15405 + 0.90723*sigmaCut - 0.23584*sigmaCut**2.0 + 0.020142*sigmaCut**3.0)

	dataSigma = dataSigma / math.sqrt(len(good)-1)

	return dataMean
	

def robustSigma(inputData):
	"""Robust estimator of the standard deviation of a data set.  Based on 
	robust_sigma function from the AstroIDL User's Library."""

	data = inputData.ravel()

	data0 = numpy.median(data)
	maxAbsDev = numpy.median(numpy.abs(data-data0)) / 0.6745
	if maxAbsDev < __epsilon:
		maxAbsDev = (numpy.abs(data-data0)).mean() / 0.8000
	if maxAbsDev < __epsilon:
		sigma = 0.0
		return sigma

	u = (data-data0) / 6.0 / maxAbsDev
	u2 = u**2.0
	good = numpy.where( u2 <= 1.0 )
	good = good[0]
	if len(good) < 3:
		print "WARNING:  Distribution is too strange to compute standard deviation"
		sigma = -1.0
		return sigma

	numerator = ((data[good]-data0)**2.0 * (1.0-u2[good])**2.0).sum()
	nElements = (data.ravel()).shape[0]
	denominator = ((1.0-u2[good])*(1.0-5.0*u2[good])).sum()
	sigma = nElements*numerator / (denominator*(denominator-1.0))
	if sigma > 0:
		sigma = math.sqrt(sigma)
	else:
		sigma = 0.0

	return sigma
